module Mgu where

open import Data.Unit.Base using (⊤; tt)
open import Data.Nat.Base
open import Data.Nat.Properties
open import Data.Nat.Properties.Simple
open import Data.Fin hiding (_+_; _≤_; _<_)
open import Data.Sum
open import Data.Product
open import Data.Maybe.Base
open import Data.Empty
open import Data.List hiding (_++_)
open import Function using (_∘_)
open import Relation.Nullary
open import Relation.Binary.PropositionalEquality
open ≡-Reasoning

open import Regular
open import Properties
open import Occur
open import Subst

---------------------------------------------------------------

sub-++-commute-fmap2 : (G : Code) → {F : Code} → {m m′ m′′ : ℕ} →
        (σ₂ : AList F m′ m′′) → (σ₁ : AList F m m′) →
        (d : ⟦ G ⟧ (μ F m)) →
        applySub-f G (σ₂ ++ σ₁) d
        ≡ applySub-f G σ₂ (applySub-f G σ₁ d)
sub-++-commute-fmap2 G σ₂ σ₁ d =
  begin
    applySub-f G (σ₂ ++ σ₁) d
  ≡⟨ sym (sub-++-commute-fmap G σ₂ σ₁ d) ⟩
    ⟪fmap⟫ G (⟪cata⟫ (sub σ₂) ∘ (sub σ₁)) d
  ≡⟨ sym (fuse-fmap G (sub σ₂) (sub σ₁) d) ⟩
    applySub-f G σ₂ (applySub-f G σ₁ d)
  ∎

ununifiable : {F : Code} → {m : ℕ} → (t₁ t₂ : μ F m) → Set
ununifiable {F} {m} t₁ t₂ =
        (l′ l′′ : ℕ) → (m≤l′′ : m ≤ l′′) → (f₀ : Fin l′′ → μ F l′) →
        ¬ (⟪cata⟫≤ f₀ m≤l′′ t₁ ≡ ⟪cata⟫≤ f₀ m≤l′′ t₂)

ununifiable-f : {F : Code} → (G : Code) → {m : ℕ} →
        (d₁ d₂ : ⟦ G ⟧ (μ F m)) → Set
ununifiable-f {F} G {m} d₁ d₂ =
        (l′ l′′ : ℕ) → (m≤l′′ : m ≤ l′′) → (f₀ : Fin l′′ → μ F l′) →
        ¬ (⟪fmap⟫≤ G f₀ m≤l′′ d₁ ≡ ⟪fmap⟫≤ G f₀ m≤l′′ d₂)

-- returns a substitution that unifies meta-variables x₁ and x₂
flexFlex : {F : Code} → (m : ℕ) → (x₁ x₂ : Fin m) →
        (Σ[ m′ ∈ ℕ ] Σ[ σ ∈ AList F m m′ ] mg ⟪ x₁ ⟫ ⟪ x₂ ⟫ (sub σ))
flexFlex zero () x₂
flexFlex (suc m) x₁ x₂ with thick₂ x₁ x₂
... | inj₁ x₁≡x₂ rewrite x₁≡x₂ = (suc m , anil , mgAnil ⟪ x₂ ⟫)
... | inj₂ (x₂′ , thinx₁x₂′≡x₂) rewrite sym thinx₁x₂′≡x₂ =
  (m , anil asnoc ⟪ x₂′ ⟫ / x₁ , mgFor x₁ ⟪ x₂′ ⟫)

private
  flexRigidBot : {F : Code} → {m m′ m′′′ : ℕ} →
          (f : Ctx⟦ F ⟧ (μ F (suc m))) →
          (fs : Context F (μ F (suc m))) →
          (x : Fin (suc m)) →
          (t : (μ F (suc m))) →
          (eq₀ : plug (f ∷ fs) ⟪ x ⟫ ≡ t) →
          (1+m≤m′′′ : suc m ≤ m′′′) →
          (f′ : Fin m′′′ → μ F m′) →
          (eq : f′ (inject≤ x 1+m≤m′′′) ≡ ⟪cata⟫≤ f′ 1+m≤m′′′ t) → ⊥
  flexRigidBot {F} f fs x t eq₀ 1+m≤m′′′ f′ eq
    with cong (⟪cata⟫≤ f′ 1+m≤m′′′) eq₀
  ... | eq₁ rewrite applySubPlug≤ f′ 1+m≤m′′′ (f ∷ fs) ⟪ x ⟫ | sym eq
    with noCycle (f′ (inject≤ x 1+m≤m′′′)) (applySubCs≤ f′ 1+m≤m′′′ (f ∷ fs)) eq₁
  ... | ()

  occurred : {F : Code} → {m : ℕ} →
          (f : Ctx⟦ F ⟧ (μ F (suc m))) →
          (fs : Context F (μ F (suc m))) →
          (x : Fin (suc m)) →
          (d : ⟦ F ⟧ (μ F (suc m))) →
          (eq : plug {F} (f ∷ fs) ⟪ x ⟫ ≡ ⟨ d ⟩) →
          ununifiable {F} ⟪ x ⟫ ⟨ d ⟩
  occurred f fs x d eq l′ l′′ = flexRigidBot f fs x ⟨ d ⟩ eq

-- returns a substitution that unifies x and a type t
-- if x occurs in t, it returns nothing
flexRigid : {F : Code} → {m : ℕ} → (x : Fin m) → (d : ⟦ F ⟧ (μ F m)) →
        ununifiable {F} ⟪ x ⟫ ⟨ d ⟩
      ⊎ (Σ[ m' ∈ ℕ ] Σ[ σ ∈ AList F m m' ] mg ⟪ x ⟫ ⟨ d ⟩ (sub σ))
flexRigid {m = zero} () d
flexRigid {m = suc m} x d with check x ⟨ d ⟩
... | inj₁ ([] , ())
... | inj₁ (f ∷ fs , eq) = inj₁ (occurred f fs x d eq) -- x occurs in ⟨ d ⟩
... | inj₂ (t′ , eq) with mgFor x t′
... | mg-t′forx rewrite eq = inj₂ (m , anil asnoc t′ / x , mg-t′forx)

private
  ununifiable-sym : {F : Code} → {m : ℕ} →
          (t₁ t₂ : μ F m) →
          ununifiable {F} t₂ t₁ → ununifiable {F} t₁ t₂
  ununifiable-sym t₁ t₂ f l′ l′′ m≤l′′ f′ eq =
    f l′ l′′ m≤l′′ f′ (sym eq)

  ununifiable-⟨d⟩ : {F : Code} → {m : ℕ} →
          (d₁ d₂ : ⟦ F ⟧ (μ F m)) →
          ununifiable-f F d₁ d₂ → ununifiable {F} ⟨ d₁ ⟩ ⟨ d₂ ⟩
  ununifiable-⟨d⟩ d₁ d₂ f l′ l′′ m≤l′′ f′ eq =
    f l′ l′′ m≤l′′ f′ (⟨-⟩-equal eq)

  ununifiable-inj₁ : {F : Code} → (G₁ G₂ : Code) → {m : ℕ} →
          (d₁ d₂ : ⟦ G₁ ⟧ (μ F m)) →
          ununifiable-f G₁ d₁ d₂ →
          ununifiable-f (G₁ ⊕ G₂) (inj₁ d₁) (inj₁ d₂)
  ununifiable-inj₁ G₁ G₂ d₁ d₂ f l′ l′′ m≤l′′ f′ eq
    = f l′ l′′ m≤l′′ f′ (inj₁-equal eq)

  ununifiable-inj₂ : {F : Code} → (G₁ G₂ : Code) → {m : ℕ} →
          (d₁ d₂ : ⟦ G₂ ⟧ (μ F m)) →
          ununifiable-f G₂ d₁ d₂ →
          ununifiable-f (G₁ ⊕ G₂) (inj₂ d₁) (inj₂ d₂)
  ununifiable-inj₂ G₁ G₂ d₁ d₂ f l′ l′′ m≤l′′ f′ eq
    = f l′ l′′ m≤l′′ f′ (inj₂-equal eq)

  ununifiable-proj₁ : {F : Code} → (G₁ G₂ : Code) → {m : ℕ} →
          (d₁₁ d₂₁ : ⟦ G₁ ⟧ (μ F m)) → (d₁₂ d₂₂ : ⟦ G₂ ⟧ (μ F m)) →
          ununifiable-f G₁ d₁₁ d₂₁ →
          ununifiable-f (G₁ ⊗ G₂) (d₁₁ , d₁₂) (d₂₁ , d₂₂)
  ununifiable-proj₁ G₁ G₂ d₁₁ d₂₁ d₁₂ d₂₂ f l′ l′′ m≤l′′ f′ eq =
    f l′ l′′ m≤l′′ f′ (proj₁ (pair-equal eq))

  mg-sym : {F : Code} → {m m′ : ℕ} →
          (t₁ t₂ : μ F m) → (f : Fin m → μ F m′) →
          mg t₂ t₁ f → mg t₁ t₂ f
  mg-sym t₁ t₂ f (eq , g) =
    (sym eq , λ l′ l′′ f′ m≤l′′ eq′ → g l′ l′′ f′ m≤l′′ (sym eq′))

  mg-⟨d⟩ : {F : Code} → {m m′ : ℕ} →
          (d₁ d₂ : ⟦ F ⟧ (μ F m)) →
          (f : Fin m → μ F m′) →
          mg-f F d₁ d₂ f → mg {F} ⟨ d₁ ⟩ ⟨ d₂ ⟩ f
  mg-⟨d⟩ d₁ d₂ f (eq , g) =
    (cong ⟨_⟩ eq , λ l′ l′′ f′ m≤l′′ eq′ → g l′ l′′ f′ m≤l′′ (⟨-⟩-equal eq′))

  mg-inj₁ : {F : Code} → (G₁ G₂ : Code) → {m m′ : ℕ} →
          (d₁ d₂ : ⟦ G₁ ⟧ (μ F m)) →
          (f : Fin m → μ F m′) →
          mg-f G₁ d₁ d₂ f → mg-f (G₁ ⊕ G₂) (inj₁ d₁) (inj₁ d₂) f
  mg-inj₁ G₁ G₂ d₁ d₂ f (eq , g) =
    (cong inj₁ eq , λ l′ l′′ f′ m≤l′′ eq′ → g l′ l′′ f′ m≤l′′ (inj₁-equal eq′))

  mg-inj₂ : {F : Code} → (G₁ G₂ : Code) → {m m′ : ℕ} →
          (d₁ d₂ : ⟦ G₂ ⟧ (μ F m)) →
          (f : Fin m → μ F m′) →
          mg-f G₂ d₁ d₂ f → mg-f (G₁ ⊕ G₂) (inj₂ d₁) (inj₂ d₂) f
  mg-inj₂ G₁ G₂ d₁ d₂ f (eq , g) =
    (cong inj₂ eq , λ l′ l′′ f′ m≤l′′ eq′ → g l′ l′′ f′ m≤l′′ (inj₂-equal eq′))

  ununifiable-proj₂ : {F : Code} → (G₁ G₂ : Code) → {m m₁ : ℕ} →
          (d₁₁ d₂₁ : ⟦ G₁ ⟧ (μ F m)) → (d₁₂ d₂₂ : ⟦ G₂ ⟧ (μ F m)) →
          (σ₁ : AList F m m₁) →
          mg-f G₁ d₁₁ d₂₁ (sub σ₁) →
          ununifiable-f G₂ (applySub-f G₂ σ₁ d₁₂) (applySub-f G₂ σ₁ d₂₂) →
          ununifiable-f (G₁ ⊗ G₂) (d₁₁ , d₁₂) (d₂₁ , d₂₂)
  ununifiable-proj₂ G₁ G₂ d₁₁ d₂₁ d₁₂ d₂₂ σ₁ (eq₁ , g₁) f₂ l′ l′′ m≤l′′ f′ eq
    with g₁ l′ l′′ f′ m≤l′′ (proj₁ (pair-equal eq))
  ... | (k′′ , m₁≤k′′ , para , f₂′ , eq₂) =
    f₂ l′ k′′ m₁≤k′′ f₂′
      (begin
         ⟪fmap⟫≤ G₂ f₂′ m₁≤k′′ (applySub-f G₂ σ₁ d₁₂)
       ≡⟨ cong (⟪fmap⟫ G₂ f₂′) (liftSubCommute-fmap G₂ σ₁ para d₁₂) ⟩
         ⟪fmap⟫ G₂ f₂′ (applySub-f≤ G₂ (liftAList≤ para σ₁) m≤l′′ d₁₂)
       ≡⟨ sym (applyUnifierAppend-f′ G₂ f₂′ σ₁ para (⟪fmap⟫ G₂ (⟪inject≤⟫ m≤l′′) d₁₂)) ⟩
         ⟪fmap⟫ G₂ (f₂′ +⟨ para ⟩′ sub σ₁) (⟪fmap⟫ G₂ (⟪inject≤⟫ m≤l′′) d₁₂)
       ≡⟨ fuse-fmap G₂ (f₂′ +⟨ para ⟩′ sub σ₁) (⟪inject≤⟫ m≤l′′) d₁₂ ⟩
         ⟪fmap⟫ G₂ (⟪cata⟫ (f₂′ +⟨ para ⟩′ sub σ₁) ∘ (⟪inject≤⟫ m≤l′′)) d₁₂
       ≡⟨ sym (cata-ext-fmap G₂ (⟪cata⟫ f′ ∘ (⟪inject≤⟫ m≤l′′))
                 (⟪cata⟫ (f₂′ +⟨ para ⟩′ sub σ₁) ∘ (⟪inject≤⟫ m≤l′′)) eq₂ d₁₂) ⟩
         ⟪fmap⟫ G₂ (⟪cata⟫ f′ ∘ (⟪inject≤⟫ m≤l′′)) d₁₂
       ≡⟨ sym (fuse-fmap G₂ f′ (⟪inject≤⟫ m≤l′′) d₁₂) ⟩
         ⟪fmap⟫≤ G₂ f′ m≤l′′ d₁₂
       ≡⟨ proj₂ (pair-equal eq) ⟩
         ⟪fmap⟫≤ G₂ f′ m≤l′′ d₂₂
       ≡⟨ fuse-fmap G₂ f′ (⟪inject≤⟫ m≤l′′) d₂₂ ⟩
         ⟪fmap⟫ G₂ (⟪cata⟫ f′ ∘ (⟪inject≤⟫ m≤l′′)) d₂₂
       ≡⟨ cata-ext-fmap G₂ (⟪cata⟫ f′ ∘ (⟪inject≤⟫ m≤l′′))
            (⟪cata⟫ (f₂′ +⟨ para ⟩′ sub σ₁) ∘ (⟪inject≤⟫ m≤l′′)) eq₂ d₂₂ ⟩
         ⟪fmap⟫ G₂ (⟪cata⟫ (f₂′ +⟨ para ⟩′ sub σ₁) ∘ (⟪inject≤⟫ m≤l′′)) d₂₂
       ≡⟨ sym (fuse-fmap G₂ (f₂′ +⟨ para ⟩′ sub σ₁) (⟪inject≤⟫ m≤l′′) d₂₂) ⟩
         ⟪fmap⟫ G₂ (f₂′ +⟨ para ⟩′ sub σ₁) (⟪fmap⟫ G₂ (⟪inject≤⟫ m≤l′′) d₂₂)
       ≡⟨ applyUnifierAppend-f′ G₂ f₂′ σ₁ para (⟪fmap⟫ G₂ (⟪inject≤⟫ m≤l′′) d₂₂) ⟩
         ⟪fmap⟫ G₂ f₂′ (applySub-f≤ G₂ (liftAList≤ para σ₁) m≤l′′ d₂₂)
       ≡⟨ sym (cong (⟪fmap⟫ G₂ f₂′) (liftSubCommute-fmap G₂ σ₁ para d₂₂)) ⟩
         ⟪fmap⟫≤ G₂ f₂′ m₁≤k′′ (applySub-f G₂ σ₁ d₂₂)
       ∎)

  composeMostGeneral : {G₁ G₂ F : Code} → {m′ m′′ m₁ m₂ m′′′ : ℕ} →
          (σ₁ : AList F m′ m₁) → (σ₂ : AList F m₁ m₂) →
          (d₁₁ d₂₁ : ⟦ G₁ ⟧ (μ F m′)) → (d₁₂ d₂₂ : ⟦ G₂ ⟧ (μ F m′)) →
          (f′ : Fin m′′′ → μ F m′′) →
          (m′≤m′′′ : m′ ≤ m′′′) →
          (eq : (fmap {F} G₁ ⟨_⟩ f′ (fmap G₁ ⟨_⟩ (λ x → ⟪ inject≤ x m′≤m′′′ ⟫)
                  d₁₁)
                ≡
                 fmap {F} G₁ ⟨_⟩ f′ (fmap G₁ ⟨_⟩ (λ x → ⟪ inject≤ x m′≤m′′′ ⟫)
                  d₂₁)) ×
                (fmap {F} G₂ ⟨_⟩ f′ (fmap G₂ ⟨_⟩ (λ x → ⟪ inject≤ x m′≤m′′′ ⟫)
                  d₁₂)
                ≡
                 fmap {F} G₂ ⟨_⟩ f′ (fmap G₂ ⟨_⟩ (λ x → ⟪ inject≤ x m′≤m′′′ ⟫)
                  d₂₂))) →
          (f₁ : (l′ l′′ : ℕ) (f′ : Fin l′′ → μ F l′) → (m′≤l′′ : m′ ≤ l′′) →
               fmap {F} G₁ ⟨_⟩ f′ (fmap G₁ ⟨_⟩ (λ x → ⟪ inject≤ x m′≤l′′ ⟫)
                 d₁₁) ≡
               fmap {F} G₁ ⟨_⟩ f′ (fmap G₁ ⟨_⟩ (λ x → ⟪ inject≤ x m′≤l′′ ⟫)
                 d₂₁) →
               Σ[ k′′ ∈ ℕ ]
               Σ[ m₁≤k′′ ∈ m₁ ≤ k′′ ]
               Σ[ para ∈ m′≤l′′ // m₁≤k′′ ]
               Σ (Fin k′′ → μ F l′)
               (λ f → (x : Fin m′) →
                f′ (inject≤ x m′≤l′′) ≡
                (f +⟨ para ⟩′ (sub σ₁)) (inject≤ x m′≤l′′))) →
          (f₂ : (l′ l′′ : ℕ) (f′ : Fin l′′ → μ F l′) → (m₁≤l′′ : m₁ ≤ l′′) →
               fmap {F} G₂ ⟨_⟩ f′ (fmap G₂ ⟨_⟩ (λ x → ⟪ inject≤ x m₁≤l′′ ⟫)
                 (fmap G₂ ⟨_⟩ (sub σ₁) d₁₂)) ≡
               fmap {F} G₂ ⟨_⟩ f′ (fmap G₂ ⟨_⟩ (λ x → ⟪ inject≤ x m₁≤l′′ ⟫)
                 (fmap G₂ ⟨_⟩ (sub σ₁) d₂₂)) →
               Σ[ k′′ ∈ ℕ ]
               Σ[ m₂≤k′′ ∈ m₂ ≤ k′′ ]
               Σ[ para ∈ m₁≤l′′ // m₂≤k′′ ]
               Σ (Fin k′′ → μ F l′)
               (λ f → (x : Fin m₁) →
                f′ (inject≤ x m₁≤l′′) ≡
                (f +⟨ para ⟩′ (sub σ₂)) (inject≤ x m₁≤l′′))) →
          Σ[ k′′ ∈ ℕ ]
          Σ[ m₂≤k′′ ∈ m₂ ≤ k′′ ]
          Σ[ para ∈ m′≤m′′′ // m₂≤k′′ ]
          Σ (Fin k′′ → μ F m′′)
          (λ f → (x : Fin m′) →
           f′ (inject≤ x m′≤m′′′) ≡
           (f +⟨ para ⟩′ (sub (σ₂ ++ σ₁))) (inject≤ x m′≤m′′′))
  composeMostGeneral {G₁} {G₂} {F} {m′ = m′} {m′′ = m′′} {m₁ = m₁} {m′′′ = m′′′} σ₁ σ₂ d₁₁ d₂₁ d₁₂ d₂₂ f′ m′≤m′′′ (eq₁′ , eq₂′) f₁ f₂
    with f₁ m′′ m′′′ f′ m′≤m′′′ eq₁′
  ... | (k₁′′ , m₁≤k₁′′ , para₁ , f₁′ , f₁′-eq)
    with f₂ m′′ k₁′′ f₁′ m₁≤k₁′′ eq₂′′
    where
     eq₂′′ : ⟪fmap⟫≤ G₂ f₁′ m₁≤k₁′′ (applySub-f G₂ σ₁ d₁₂) ≡
            ⟪fmap⟫≤ G₂ f₁′ m₁≤k₁′′ (applySub-f G₂ σ₁ d₂₂)
     eq₂′′ = begin
             ⟪fmap⟫≤ G₂ f₁′ m₁≤k₁′′ (applySub-f G₂ σ₁ d₁₂)
           ≡⟨ cong (⟪fmap⟫ G₂ f₁′) (liftSubCommute-fmap G₂ σ₁ para₁ d₁₂) ⟩
             ⟪fmap⟫ G₂ f₁′ (applySub-f≤ G₂ (liftAList≤ para₁ σ₁) m′≤m′′′ d₁₂)
           ≡⟨ sym (applyUnifierAppend-f′ G₂ f₁′ σ₁ para₁
                    (⟪fmap⟫ G₂ (⟪inject≤⟫ m′≤m′′′) d₁₂)) ⟩
             ⟪fmap⟫≤ G₂ (f₁′ +⟨ para₁ ⟩′ sub σ₁) m′≤m′′′ d₁₂
           ≡⟨ fuse-fmap G₂ (f₁′ +⟨ para₁ ⟩′ sub σ₁) (⟪inject≤⟫ m′≤m′′′) d₁₂ ⟩
             ⟪fmap⟫ G₂ (⟪cata⟫ (f₁′ +⟨ para₁ ⟩′ sub σ₁) ∘ (⟪inject≤⟫ m′≤m′′′)) d₁₂
           ≡⟨ sym (cata-ext-fmap G₂ (⟪cata⟫ f′ ∘ (⟪inject≤⟫ m′≤m′′′)) (⟪cata⟫ (f₁′ +⟨ para₁ ⟩′ sub σ₁) ∘ (⟪inject≤⟫ m′≤m′′′)) f₁′-eq d₁₂) ⟩
             ⟪fmap⟫ G₂ (⟪cata⟫ f′ ∘ (⟪inject≤⟫ m′≤m′′′)) d₁₂
           ≡⟨ sym (fuse-fmap G₂ f′ (⟪inject≤⟫ m′≤m′′′) d₁₂) ⟩
             ⟪fmap⟫≤ G₂ f′ m′≤m′′′ d₁₂
           ≡⟨ eq₂′ ⟩
             ⟪fmap⟫≤ G₂ f′ m′≤m′′′ d₂₂
           ≡⟨ fuse-fmap G₂ f′ (⟪inject≤⟫ m′≤m′′′) d₂₂ ⟩
             ⟪fmap⟫ G₂ (⟪cata⟫ f′ ∘ (⟪inject≤⟫ m′≤m′′′)) d₂₂
           ≡⟨ cata-ext-fmap G₂ (⟪cata⟫ f′ ∘ (⟪inject≤⟫ m′≤m′′′)) (⟪cata⟫ (f₁′ +⟨ para₁ ⟩′ sub σ₁) ∘ (⟪inject≤⟫ m′≤m′′′)) f₁′-eq d₂₂ ⟩
             ⟪fmap⟫ G₂ (⟪cata⟫ (f₁′ +⟨ para₁ ⟩′ sub σ₁) ∘ (⟪inject≤⟫ m′≤m′′′)) d₂₂
           ≡⟨ sym (fuse-fmap G₂ (f₁′ +⟨ para₁ ⟩′ sub σ₁) (⟪inject≤⟫ m′≤m′′′) d₂₂) ⟩
             ⟪fmap⟫≤ G₂ (f₁′ +⟨ para₁ ⟩′ sub σ₁) m′≤m′′′ d₂₂
           ≡⟨ applyUnifierAppend-f′ G₂ f₁′ σ₁ para₁
                (⟪fmap⟫ G₂ (⟪inject≤⟫ m′≤m′′′) d₂₂) ⟩
             ⟪fmap⟫ G₂ f₁′ (applySub-f≤ G₂ (liftAList≤ para₁ σ₁) m′≤m′′′ d₂₂)
           ≡⟨ sym (cong (⟪fmap⟫ G₂ f₁′) (liftSubCommute-fmap G₂ σ₁ para₁ d₂₂)) ⟩
             ⟪fmap⟫≤ G₂ f₁′ m₁≤k₁′′ (applySub-f G₂ σ₁ d₂₂)
           ∎
  ... | (k₂′′ , m₂≤k₂′′ , para₂ , f₂′ , f₂′-eq) =
    (k₂′′ , m₂≤k₂′′ , para-append para₁ para₂ , f₂′ ,
     λ x →
       begin
         f′ (inject≤ x m′≤m′′′)
       ≡⟨ f₁′-eq x ⟩
         applyUnifier (f₁′ +⟨ para₁ ⟩′ sub σ₁)
          (lift≤ m′≤m′′′ ⟪ x ⟫)
       ≡⟨ applyUnifierAppend′ f₁′ σ₁ para₁ (lift≤ m′≤m′′′ ⟪ x ⟫) ⟩
         applyUnifier f₁′
          (applySub (liftAList≤ para₁ σ₁)
           (lift≤ m′≤m′′′ ⟪ x ⟫))
       ≡⟨ cong (applyUnifier f₁′)
               (sym (liftSubCommute σ₁ para₁ ⟪ x ⟫)) ⟩
         applyUnifier f₁′
          (lift≤ m₁≤k₁′′ (applySub σ₁ ⟪ x ⟫))
       ≡⟨ cata-ext′′ f₁′ (applyUnifier f₂′ ∘ liftFun para₂ (sub σ₂)) m₁≤k₁′′ m₁≤k₁′′ f₂′-eq (applySub σ₁ ⟪ x ⟫) ⟩ -- meaning of f₂′-eq !!!
         applyUnifier (f₂′ +⟨ para₂ ⟩′ sub σ₂)
          (lift≤ m₁≤k₁′′ (applySub σ₁ ⟪ x ⟫))
       ≡⟨ cong (applyUnifier (f₂′ +⟨ para₂ ⟩′ sub σ₂))
           (liftSubCommute σ₁ para₁ ⟪ x ⟫) ⟩
         applyUnifier (f₂′ +⟨ para₂ ⟩′ sub σ₂)
          (applySub (liftAList≤ para₁ σ₁)
            (lift≤ m′≤m′′′ ⟪ x ⟫))
       ≡⟨ sym (applyUnifierAppend′ (f₂′ +⟨ para₂ ⟩′ sub σ₂) σ₁ para₁ (lift≤ m′≤m′′′ ⟪ x ⟫)) ⟩
         applyUnifier ((f₂′ +⟨ para₂ ⟩′ sub σ₂) +⟨ para₁ ⟩′ sub σ₁)
          (lift≤ m′≤m′′′ ⟪ x ⟫)
       ≡⟨ refl ⟩
         ((f₂′ +⟨ para₂ ⟩′ sub σ₂) +⟨ para₁ ⟩′ sub σ₁) (inject≤ x m′≤m′′′)
       ≡⟨ para-sub-append f₂′ para₂ σ₂ para₁ σ₁ x ⟩
         (f₂′ +⟨ para-append para₁ para₂ ⟩′ sub (σ₂ ++ σ₁)) (inject≤ x m′≤m′′′)
       ∎)

  mg-pair : {F : Code} → (G₁ G₂ : Code) → {m m₁ m₂ : ℕ} →
          (d₁₁ d₂₁ : ⟦ G₁ ⟧ (μ F m)) → (d₁₂ d₂₂ : ⟦ G₂ ⟧ (μ F m)) →
          (σ₁ : AList F m m₁) →
          (σ₂ : AList F m₁ m₂) →
          mg-f G₁ d₁₁ d₂₁ (sub σ₁) →
          mg-f G₂ (applySub-f G₂ σ₁ d₁₂) (applySub-f G₂ σ₁ d₂₂) (sub σ₂) →
          mg-f (G₁ ⊗ G₂) (d₁₁ , d₁₂) (d₂₁ , d₂₂) (sub (σ₂ ++ σ₁))
  mg-pair G₁ G₂ d₁₁ d₂₁ d₁₂ d₂₂ σ₁ σ₂ (eq₁ , g₁) (eq₂ , g₂)
    rewrite sub-++-commute-fmap2 G₁ σ₂ σ₁ d₁₁
          | sub-++-commute-fmap2 G₁ σ₂ σ₁ d₂₁
          | sub-++-commute-fmap2 G₂ σ₂ σ₁ d₁₂
          | sub-++-commute-fmap2 G₂ σ₂ σ₁ d₂₂
    = (cong₂ _,_ (cong (applySub-f G₁ σ₂) eq₁) eq₂ ,
       λ l′ l′′ f m≤l′′ eq →
         composeMostGeneral {G₁} {G₂} σ₁ σ₂ d₁₁ d₂₁ d₁₂ d₂₂ f m≤l′′ (pair-equal eq) g₁ g₂)

  fmap-id : (F : Code) → {m : ℕ} → (d : ⟦ F ⟧ (μ F m)) →
        fmap F ⟨_⟩ ⟪_⟫ d ≡ d
  fmap-id F d = ⟨-⟩-equal eq
    where eq : ⟪cata⟫ {F} ⟪_⟫ ⟨ d ⟩ ≡ ⟨ d ⟩
          eq = ⟪-⟫-id ⟨ d ⟩

-- returns a substitution that unifies t1 and t2 with correctness proof using accumulator
mutual
  amguAnil : {F : Code} → {m : ℕ} → (t₁ t₂ : μ F m) →
          ununifiable t₁ t₂
        ⊎ (Σ[ m′′ ∈ ℕ ] Σ[ σ ∈ AList F m m′′ ] mg t₁ t₂ (sub σ))
  amguAnil {F} ⟨ d₁ ⟩ ⟨ d₂ ⟩ with amgu′ F d₁ d₂ anil
  ... | inj₁ f rewrite fmap-id F d₁ | fmap-id F d₂
    = inj₁ (ununifiable-⟨d⟩ d₁ d₂ f)
  ... | inj₂ (m′ , σ′ , mgσ′) rewrite fmap-id F d₁ | fmap-id F d₂
    = inj₂ (m′ , σ′ , mg-⟨d⟩ d₁ d₂ (sub σ′) mgσ′)
  amguAnil ⟨ d₁ ⟩ ⟪ x₂ ⟫ with flexRigid x₂ d₁
  ... | inj₁ f = inj₁ (ununifiable-sym ⟨ d₁ ⟩ ⟪ x₂ ⟫ f)
  ... | inj₂ (m′ , σ′ , mgσ′) =
    inj₂ (m′ , σ′ , mg-sym ⟨ d₁ ⟩ ⟪ x₂ ⟫ (sub σ′) mgσ′)
  amguAnil ⟪ x₁ ⟫ ⟨ d₂ ⟩ with flexRigid x₁ d₂
  ... | inj₁ f = inj₁ f
  ... | inj₂ (m′ , σ′ , mgσ′) = inj₂ (m′ , σ′ , mgσ′)
  amguAnil {m = m} ⟪ x₁ ⟫ ⟪ x₂ ⟫ with flexFlex m x₁ x₂
  ... | (m′ , σ′ , mgσ′) = inj₂ (m′ , σ′ , mgσ′)

  amgu : {F : Code} → {m m′ : ℕ} → (t₁ t₂ : μ F m) →
          (ρ : AList F m m′) →
          ununifiable (applySub ρ t₁) (applySub ρ t₂)
        ⊎ (Σ[ m′′ ∈ ℕ ] Σ[ σ ∈ AList F m′ m′′ ]
           mg (applySub ρ t₁) (applySub ρ t₂) (sub σ))
  amgu t₁ t₂ anil
    rewrite ⟪-⟫-id t₁ | ⟪-⟫-id t₂ = amguAnil t₁ t₂
  amgu t₁ t₂ (ρ asnoc t / x)
    with amgu (⟪cata⟫ (t for x) t₁) (⟪cata⟫ (t for x) t₂) ρ
  ... | inj₁ f
    rewrite fuse (sub ρ) (t for x) t₁ | fuse (sub ρ) (t for x) t₂
    = inj₁ f
  ... | inj₂ (m′′ , σ′ , mgσ′)
    rewrite fuse (sub ρ) (t for x) t₁ | fuse (sub ρ) (t for x) t₂
    = inj₂ (m′′ , σ′ , mgσ′)

  amgu′ : {F : Code} → (G : Code) → {m : ℕ} →
          (d₁ d₂ : ⟦ G ⟧ (μ F m)) →
          {m′ : ℕ} → (ρ : AList F m m′) →
          ununifiable-f G (applySub-f G ρ d₁) (applySub-f G ρ d₂)
       ⊎ (Σ[ m′′ ∈ ℕ ] Σ[ σ ∈ AList F m′ m′′ ]
          mg-f G (applySub-f G ρ d₁) (applySub-f G ρ d₂) (sub σ))
  amgu′ U tt tt {m′} ρ = inj₂ (m′ , anil , mgAnil-f U tt)
  amgu′ I d₁ d₂ ρ = amgu d₁ d₂ ρ
  amgu′ (G₁ ⊕ G₂) (inj₁ d₁) (inj₁ d₂) ρ with amgu′ G₁ d₁ d₂ ρ
  ... | inj₁ f =
    inj₁ (ununifiable-inj₁ G₁ G₂ (applySub-f G₁ ρ d₁) (applySub-f G₁ ρ d₂) f)
  ... | inj₂ (m′′ , σ′ , mgσ′) =
    inj₂ (m′′ , σ′ , mg-inj₁ G₁ G₂ (applySub-f G₁ ρ d₁)
                                 (applySub-f G₁ ρ d₂) (sub σ′) mgσ′)
  amgu′ (G₁ ⊕ G₂) (inj₁ d₁) (inj₂ d₂) ρ = inj₁ (λ m′′ m′′′ m′≤m′′′ f′ ())
  amgu′ (G₁ ⊕ G₂) (inj₂ d₁) (inj₁ d₂) ρ = inj₁ (λ m′′ m′′′ m′≤m′′′ f′ ())
  amgu′ (G₁ ⊕ G₂) (inj₂ d₁) (inj₂ d₂) ρ with amgu′ G₂ d₁ d₂ ρ
  ... | inj₁ f =
    inj₁ (ununifiable-inj₂ G₁ G₂ (applySub-f G₂ ρ d₁) (applySub-f G₂ ρ d₂) f)
  ... | inj₂ (m′′ , σ′ , mgσ′) =
    inj₂ (m′′ , σ′ , mg-inj₂ G₁ G₂ (applySub-f G₂ ρ d₁)
                                 (applySub-f G₂ ρ d₂) (sub σ′) mgσ′)
  amgu′ {F} (G₁ ⊗ G₂) (d₁₁ , d₁₂) (d₂₁ , d₂₂) ρ
    with amgu′ G₁ d₁₁ d₂₁ ρ
  ... | inj₁ f =
    inj₁ (ununifiable-proj₁ G₁ G₂ (applySub-f G₁ ρ d₁₁) (applySub-f G₁ ρ d₂₁)
                                (applySub-f G₂ ρ d₁₂) (applySub-f G₂ ρ d₂₂) f)
  ... | inj₂ (m₁ , σ₁ , mgσ₁)
    with amgu′ G₂ d₁₂ d₂₂ (σ₁ ++ ρ)
  ... | inj₁ f
    rewrite sub-++-commute-fmap2 G₂ σ₁ ρ d₁₂
          | sub-++-commute-fmap2 G₂ σ₁ ρ d₂₂
    = inj₁ (ununifiable-proj₂ G₁ G₂ (applySub-f G₁ ρ d₁₁) (applySub-f G₁ ρ d₂₁) (applySub-f G₂ ρ d₁₂) (applySub-f G₂ ρ d₂₂) σ₁ mgσ₁ f)
  ... | inj₂ (m₂ , σ₂ , mgσ₂)
    rewrite sub-++-commute-fmap2 G₂ σ₁ ρ d₁₂
          | sub-++-commute-fmap2 G₂ σ₁ ρ d₂₂
    = inj₂ (m₂ , σ₂ ++ σ₁ , mg-pair G₁ G₂ (applySub-f G₁ ρ d₁₁) (applySub-f G₁ ρ d₂₁) (applySub-f G₂ ρ d₁₂) (applySub-f G₂ ρ d₂₂) σ₁ σ₂ mgσ₁ mgσ₂)

-- returns a substitution that unifies t₁ and t₂ with correctness proof
mgu : {F : Code} → {m : ℕ} → (t₁ t₂ : μ F m) →
       ununifiable t₁ t₂
     ⊎ (Σ[ m′ ∈ ℕ ] Σ[ σ ∈ AList F m m′ ] mg t₁ t₂ (sub σ))
mgu t₁ t₂ with amgu t₁ t₂ anil
... | inj₁ f rewrite ⟪-⟫-id t₁ | ⟪-⟫-id t₂ = inj₁ f
... | inj₂ (m′ , σ , mgσ) rewrite ⟪-⟫-id t₁ | ⟪-⟫-id t₂ = inj₂ (m′ , σ , mgσ)
