module Properties where

open import Data.Unit.Base using (⊤; tt)
open import Data.Nat.Base hiding (erase)
open import Data.Nat.Properties using (n≤m+n)
open import Data.Fin hiding (_+_; _≤_; _<_)
open import Data.Fin.Properties
open import Data.Sum
open import Data.Product
open import Function using (_∘_)
open import Relation.Binary.PropositionalEquality
open ≡-Reasoning

open import Regular

--------------------------------------------------------------------------------

inject≤-id : {m : ℕ} → (x : Fin m) → (m≤m : m ≤ m) →
         inject≤ x m≤m ≡ x
inject≤-id zero (s≤s m≤m) = refl
inject≤-id (suc x) (s≤s m≤m) = cong suc (inject≤-id x m≤m)

inject≤-trans : {l m n : ℕ} → (y : Fin l) →
        (l≤m : l ≤ m) → (m≤n : m ≤ n) → (l≤n : l ≤ n) →
        inject≤ (inject≤ y l≤m) m≤n ≡ inject≤ y l≤n
inject≤-trans {zero} () l≤m m≤n l≤n
inject≤-trans {suc l} {zero} y () m≤n l≤n
inject≤-trans {suc l} {suc m} {zero} y l≤m () l≤n
inject≤-trans {suc l} {suc m} {suc n} zero (s≤s l≤m) (s≤s m≤n) (s≤s l≤n) = refl
inject≤-trans {suc l} {suc m} {suc n} (suc y) (s≤s l≤m) (s≤s m≤n) (s≤s l≤n) = cong suc (inject≤-trans y l≤m m≤n l≤n)

--

mutual
  cata-ext-fmap : {F : Code} → (G : Code) → {m m′ : ℕ} →
          (f₁ f₂ : Fin m → μ F m′) →
          (eq : (x : Fin m) → f₁ x ≡ f₂ x) →
          (d : ⟦ G ⟧ (μ F m)) →
          fmap G ⟨_⟩ f₁ d ≡ fmap G ⟨_⟩ f₂ d
  cata-ext-fmap U f₁ f₂ eq tt = refl
  cata-ext-fmap I f₁ f₂ eq d = cata-ext f₁ f₂ eq d
  cata-ext-fmap (G₁ ⊕ G₂) f₁ f₂ eq (inj₁ d) = cong inj₁ (cata-ext-fmap G₁ f₁ f₂ eq d)
  cata-ext-fmap (G₁ ⊕ G₂) f₁ f₂ eq (inj₂ d) = cong inj₂ (cata-ext-fmap G₂ f₁ f₂ eq d)
  cata-ext-fmap (G₁ ⊗ G₂) f₁ f₂ eq (d₁ , d₂) =
    cong₂ _,_ (cata-ext-fmap G₁ f₁ f₂ eq d₁) (cata-ext-fmap G₂ f₁ f₂ eq d₂)

  cata-ext : {F : Code} → {m m′ : ℕ} →
          (f₁ f₂ : Fin m → μ F m′) →
          (eq : (x : Fin m) → f₁ x ≡ f₂ x) →
          (t : μ F m) → cata ⟨_⟩ f₁ t ≡ cata ⟨_⟩ f₂ t
  cata-ext {F} f₁ f₂ eq ⟨ x ⟩ = cong ⟨_⟩ (cata-ext-fmap F f₁ f₂ eq x)
  cata-ext {F} f₁ f₂ eq ⟪ x ⟫ = eq x

private
  ⟨cata-ext⟩′′ : {F : Code} → (G : Code) → {m m′ m₁′′ m₂′′ : ℕ} →
          (f₁ : Fin m₁′′ → μ F m′) →
          (f₂ : Fin m₂′′ → μ F m′) →
          (m≤m₁′′ : m ≤ m₁′′) →
          (m≤m₂′′ : m ≤ m₂′′) →
          (d : ⟦ G ⟧ (μ F m)) →
          (r : ⟦ G ⟧′ (λ t → cata ⟨_⟩ f₁ (lift≤ m≤m₁′′ t) ≡
                              cata ⟨_⟩ f₂ (lift≤ m≤m₂′′ t)) d) →
          fmap {F} G ⟨_⟩ f₁ (fmap G ⟨_⟩ (λ x → ⟪ inject≤ x m≤m₁′′ ⟫) d) ≡
          fmap {F} G ⟨_⟩ f₂ (fmap G ⟨_⟩ (λ x → ⟪ inject≤ x m≤m₂′′ ⟫) d)
  ⟨cata-ext⟩′′ U f₁ f₂ m≤m₁′′ m≤m₂′′ tt tt = refl
  ⟨cata-ext⟩′′ I f₁ f₂ m≤m₁′′ m≤m₂′′ d r = r
  ⟨cata-ext⟩′′ (G₁ ⊕ G₂) f₁ f₂ m≤m₁′′ m≤m₂′′ (inj₁ d) r =
    cong inj₁ (⟨cata-ext⟩′′ G₁ f₁ f₂ m≤m₁′′ m≤m₂′′ d r)
  ⟨cata-ext⟩′′ (G₁ ⊕ G₂) f₁ f₂ m≤m₁′′ m≤m₂′′ (inj₂ d) r =
    cong inj₂ (⟨cata-ext⟩′′ G₂ f₁ f₂ m≤m₁′′ m≤m₂′′ d r)
  ⟨cata-ext⟩′′ (G₁ ⊗ G₂) f₁ f₂ m≤m₁′′ m≤m₂′′ (d₁ , d₂) (r₁ , r₂) =
    cong₂ _,_ (⟨cata-ext⟩′′ G₁ f₁ f₂ m≤m₁′′ m≤m₂′′ d₁ r₁)
              (⟨cata-ext⟩′′ G₂ f₁ f₂ m≤m₁′′ m≤m₂′′ d₂ r₂)

cata-ext′′ : {F : Code} → {m m′ m₁′′ m₂′′ : ℕ} →
        (f₁ : Fin m₁′′ → μ F m′) →
        (f₂ : Fin m₂′′ → μ F m′) →
        (m≤m₁′′ : m ≤ m₁′′) →
        (m≤m₂′′ : m ≤ m₂′′) →
        (eq : (x : Fin m) →
              f₁ (inject≤ x m≤m₁′′) ≡ f₂ (inject≤ x m≤m₂′′)) →
        (t : μ F m) →
        cata ⟨_⟩ f₁ (lift≤ m≤m₁′′ t) ≡ cata ⟨_⟩ f₂ (lift≤ m≤m₂′′ t)
cata-ext′′ {F} {m} f₁ f₂ m≤m₁′′ m≤m₂′′ eq t =
  ind (λ t → cata ⟨_⟩ f₁ (lift≤ m≤m₁′′ t) ≡
               cata ⟨_⟩ f₂ (lift≤ m≤m₂′′ t))
      (λ d r → cong ⟨_⟩ (⟨cata-ext⟩′′ F f₁ f₂ m≤m₁′′ m≤m₂′′ d r))
      (λ x → eq x)
      t

-- fusion of functions on meta-variables

mutual
  fuse-fmap : (G : Code) → {F : Code} → {m₁ m₂ m₃ : ℕ} →
          (f₂ : Fin m₂ → μ F m₃) → (f₁ : Fin m₁ → μ F m₂) →
          (d : ⟦ G ⟧ (μ F m₁)) →
          fmap G ⟨_⟩ f₂ (fmap G ⟨_⟩ f₁ d) ≡ fmap G ⟨_⟩ (cata ⟨_⟩ f₂ ∘ f₁) d
  fuse-fmap U f₂ f₁ tt = refl
  fuse-fmap I f₂ f₁ d = fuse f₂ f₁ d
  fuse-fmap (G₁ ⊕ G₂) f₂ f₁ (inj₁ d) = cong inj₁ (fuse-fmap G₁ f₂ f₁ d)
  fuse-fmap (G₁ ⊕ G₂) f₂ f₁ (inj₂ d) = cong inj₂ (fuse-fmap G₂ f₂ f₁ d)
  fuse-fmap (G₁ ⊗ G₂) f₂ f₁ (d₁ , d₂) =
    cong₂ _,_ (fuse-fmap G₁ f₂ f₁ d₁) (fuse-fmap G₂ f₂ f₁ d₂)

  fuse : {F : Code} → {m₁ m₂ m₃ : ℕ} →
          (f₂ : Fin m₂ → μ F m₃) → (f₁ : Fin m₁ → μ F m₂) →
          (t : μ F m₁) →
          cata ⟨_⟩ f₂ (cata ⟨_⟩ f₁ t) ≡ cata ⟨_⟩ (cata ⟨_⟩ f₂ ∘ f₁) t
  fuse {F} f₂ f₁ ⟨ x ⟩ = cong ⟨_⟩ (fuse-fmap F f₂ f₁ x)
  fuse {F} f₂ f₁ ⟪ x ⟫ = refl

fuse-lift : {F : Code} → {l m n : ℕ} →
        (l≤m : l ≤ m) → (m≤n : m ≤ n) → (l≤n : l ≤ n) →
        (t : μ F l) →
        lift≤ m≤n (lift≤ l≤m t) ≡ lift≤ l≤n t
fuse-lift l≤m m≤n l≤n t =
  begin
    cata ⟨_⟩ (⟪inject≤⟫ m≤n) (cata ⟨_⟩ (⟪inject≤⟫ l≤m) t)
  ≡⟨ fuse (⟪inject≤⟫ m≤n) (⟪inject≤⟫ l≤m) t ⟩
    cata ⟨_⟩ (λ x → ⟪ inject≤ (inject≤ x l≤m) m≤n ⟫) t
  ≡⟨ cata-ext (λ x → ⟪ inject≤ (inject≤ x l≤m) m≤n ⟫) (⟪inject≤⟫ l≤n)
              (λ x → cong ⟪_⟫ (inject≤-trans x l≤m m≤n l≤n)) t ⟩
    cata ⟨_⟩ (⟪inject≤⟫ l≤n) t
  ∎

private
  ⟨cata-id⟩ : {F : Code} → (G : Code) → {m : ℕ} →
            (⟪id⟫ : Fin m → μ F m) →
            (d : ⟦ G ⟧ (μ F m)) →
            (x : ⟦ G ⟧′ (λ t → cata ⟨_⟩ ⟪id⟫ t ≡ t) d) →
            fmap G ⟨_⟩ ⟪id⟫ d ≡ d
  ⟨cata-id⟩ U ⟪id⟫ tt tt = refl
  ⟨cata-id⟩ I ⟪id⟫ d x = x
  ⟨cata-id⟩ (G₁ ⊕ G₂) ⟪id⟫ (inj₁ d) x = cong inj₁ (⟨cata-id⟩ G₁ ⟪id⟫ d x)
  ⟨cata-id⟩ (G₁ ⊕ G₂) ⟪id⟫ (inj₂ d) x = cong inj₂ (⟨cata-id⟩ G₂ ⟪id⟫ d x)
  ⟨cata-id⟩ (G₁ ⊗ G₂) ⟪id⟫ (d₁ , d₂) (x₁ , x₂) =
    cong₂ _,_ (⟨cata-id⟩ G₁ ⟪id⟫ d₁ x₁) (⟨cata-id⟩ G₂ ⟪id⟫ d₂ x₂)

cata-id : {F : Code} → {m : ℕ} →
        (⟪id⟫ : Fin m → μ F m) →
        (⟪id⟫-eq : (x : Fin m) → ⟪id⟫ x ≡ ⟪ x ⟫) →
        (t : μ F m) → cata ⟨_⟩ ⟪id⟫ t ≡ t
cata-id {F} ⟪id⟫ ⟪id⟫-eq t =
  ind (λ t → cata ⟨_⟩ ⟪id⟫ t ≡ t)
      (λ d x → cong ⟨_⟩ (⟨cata-id⟩ F ⟪id⟫ d x))
      ⟪id⟫-eq
      t

⟪-⟫-id : {F : Code} → {m : ℕ} → (t : μ F m) → ⟪cata⟫ ⟪_⟫ t ≡ t
⟪-⟫-id t = cata-id ⟪_⟫ (λ x → refl) t

lift≤-id : {F : Code} → {m : ℕ} → (m≤m : m ≤ m) →
            (t : μ F m) → lift≤ m≤m t ≡ t
lift≤-id m≤m t =
  cata-id (λ x → ⟪ inject≤ x m≤m ⟫)
          (λ x → cong ⟪_⟫ (inject≤-id x m≤m))
          t
