module Subst where

open import Data.Unit.Base using (⊤; tt)
open import Data.Nat
open import Data.Nat.Properties
open import Data.Fin hiding (_+_; _≤_; _<_)
open import Data.Fin.Properties
open import Data.Sum
open import Data.Product
open import Data.Maybe.Base
open import Data.Empty
open import Data.List hiding (_++_; reverse)
open import Function using (_∘_)
open import Relation.Nullary
open import Relation.Binary using (DecTotalOrder)
open import Relation.Binary.PropositionalEquality
open ≡-Reasoning

open import Regular
open import Properties
open import Occur

---------------------------------------------------------------

≤-trans : ∀ {m j k} → (m≤j : m ≤ j) → (j≤k : j ≤ k) → m ≤ k
≤-trans = DecTotalOrder.trans decTotalOrder

private
  thick-max : {l l′ : ℕ} →
          (y : Fin (suc l)) → (l≤l′ : l ≤ l′) → (l≤1+l′ : l ≤ suc l′) →
          thick (inject≤ y (s≤s l≤1+l′)) (suc (fromℕ l′)) ≡ just (fromℕ l′)
  thick-max zero l≤l′ l≤1+l′ = refl
  thick-max {zero} (suc ()) l≤l′ l≤1+l′
  thick-max {suc l} (suc y) (s≤s l≤l′) (s≤s l≤1+l′)
    rewrite thick-max y l≤l′ l≤1+l′ = refl

-- private
data FinOrdering {l : ℕ} : Fin (suc l) → Set where
  max : FinOrdering (fromℕ l)
  less : (x : Fin l) → FinOrdering (inject≤ x (n≤1+n l))

finCompare : {l : ℕ} → (x : Fin (suc l)) → FinOrdering x
finCompare {zero} zero = max
finCompare {zero} (suc ())
finCompare {suc l} zero = less (zero {l})
finCompare {suc l} (suc x) with finCompare x
finCompare {suc l} (suc .(fromℕ l)) | max = max
finCompare {suc l} (suc .(inject≤ x (n≤1+n l))) | less x = less (suc x)

fin-inv : {l : ℕ} → (x : Fin (suc l)) →
        x ≡ fromℕ l ⊎ Σ[ x′ ∈ Fin l ] x ≡ inject≤ x′ (n≤1+n l)
fin-inv {l} x with finCompare x
fin-inv {l} .(fromℕ l) | max = inj₁ refl
fin-inv {l} .(inject≤ {l} x (n≤1+n l)) | less x = inj₂ (x , refl)

-- t for x : returns a unifier that maps x to t
_for_ : {F : Code} → {m : ℕ} →
        (t : μ F m) → (x : Fin (suc m)) → Fin (suc m) → μ F m
(t for x) y with thick x y
... | nothing = t -- x = y
... | just y′ = ⟪ y′ ⟫ -- y is thickened to y′

for-id : {F : Code} → {m : ℕ} → (t : μ F m) → (x : Fin (suc m)) →
        (t for x) x ≡ t
for-id t x rewrite thickxx≡nothing x = refl

-- AList F l m : substitution from l meta-variables to m meta-variables
data AList (F : Code) : (l m : ℕ) → Set where
  anil : {m : ℕ} → AList F m m -- empty substitution
  _asnoc_/_ : {l m : ℕ} → (σ : AList F l m) → (t : μ F l) →
    (x : Fin (suc l)) → AList F (suc l) m -- maps x to t before σ

-- AList F m n induces inequality n ≤′ m
AListLeq : {F : Code} → {m n : ℕ} → AList F m n → n ≤′ m
AListLeq {m = m} anil = ≤′-refl
AListLeq (σ asnoc t / x) with AListLeq σ
... | m'≤m = ≤′-step m'≤m

-- private
mutual
  AListInv : {F : Code} → {m n : ℕ} → (σ : AList F m n) →
          (Σ[ m≡n ∈ m ≡ n ]
           σ ≡ subst (λ l → AList F m l) m≡n (anil {m = m})) ⊎
          (Σ[ m' ∈ ℕ ] Σ[ sucm'≡m ∈ suc m' ≡ m ]
           Σ[ σ' ∈ AList F m' n ] Σ[ t ∈ μ F m' ] Σ[ x ∈ Fin (suc m') ]
           σ ≡ subst (λ l → AList F l n) sucm'≡m (σ' asnoc t / x))
  AListInv anil = inj₁ (refl , refl)
  AListInv {m = suc m} (σ asnoc t / x) = inj₂ (m , refl , σ , t , x , refl)

  AListInvAsnoc : {F : Code} → {m n : ℕ} → (n<m : n < m) → (σ : AList F m n) →
          Σ[ m' ∈ ℕ ] Σ[ sucm'≡m ∈ suc m' ≡ m ]
          Σ[ σ' ∈ AList F m' n ] Σ[ t ∈ μ F m' ] Σ[ x ∈ Fin (suc m') ]
          σ ≡ subst (λ l → AList F l n) sucm'≡m (σ' asnoc t / x)
  AListInvAsnoc {m = m} m<m anil with 1+n≰n m<m 
  ... | ()
  AListInvAsnoc {m = suc m} n<sucm (σ asnoc t / x)
    = (m , refl , σ , t , x , refl)

  AListInvAsnoc′ : {F : Code} → {m n : ℕ} → (n≤m : n ≤ m) → (σ : AList F (suc m) n) →
          Σ[ σ' ∈ AList F m n ] Σ[ t ∈ μ F m ] Σ[ x ∈ Fin (suc m) ]
          σ ≡ σ' asnoc t / x
  AListInvAsnoc′ n≤m anil with 1+n≰n n≤m 
  ... | ()
  AListInvAsnoc′ n≤m (σ asnoc t / x)
    = (σ , t , x , refl)

  AListInvAnil : {F : Code} → {m : ℕ} → (σ : AList F m m) → σ ≡ anil
  AListInvAnil anil = refl
  AListInvAnil (σ asnoc t / x) with 1+n≰n (≤′⇒≤ (AListLeq σ))
  ... | ()

-- append two substitutions
_++_ : {F : Code} → {l m n : ℕ} →
       (ρ : AList F m n) → (σ : AList F l m) → AList F l n
ρ ++ anil = ρ
ρ ++ (σ asnoc t / x) = (ρ ++ σ) asnoc t / x

data _//_ : {l l′ m m′ : ℕ} → l ≤ l′ → m ≤ m′ → Set where
  Refl : {m m′ : ℕ} → (m≤m′ : m ≤ m′) → m≤m′ // m≤m′
  Step : {l l′ m m′ : ℕ} → {l≤l′ : l ≤ l′} →
         {m≤m′ : m ≤ m′} → l≤l′ // m≤m′ → s≤s l≤l′ // m≤m′

para-append : {l₁ l₂ m₁ m₂ n₁ n₂ : ℕ} →
        {l₁≤l₂ : l₁ ≤ l₂} → {m₁≤m₂ : m₁ ≤ m₂} → {n₁≤n₂ : n₁ ≤ n₂} →
        (para₁ : l₁≤l₂ // m₁≤m₂) → 
        (para₂ : m₁≤m₂ // n₁≤n₂) → 
        l₁≤l₂ // n₁≤n₂
para-append (Refl l₁≤l₂) para₂ = para₂
para-append (Step para₁) para₂ = Step (para-append para₁ para₂)

data _//′_ : {l l′ m m′ : ℕ} → m ≤ l → m′ ≤ l′ → Set where
  Refl′ : {m l : ℕ} → (m≤l : m ≤ l) → m≤l //′ m≤l
  Step′ : {l l′ m m′ : ℕ} → {m≤l : m ≤ l} →
         {m′≤l′ : m′ ≤ l′} → m≤l //′ m′≤l′ → m≤l //′ s≤s m′≤l′

private
  s≤s//′ : {l l′ m m′ : ℕ} → {m≤l : m ≤ l} → {m′≤l′ : m′ ≤ l′} →
          m≤l //′ m′≤l′ → s≤s m≤l //′ s≤s m′≤l′
  s≤s//′ (Refl′ l≤l′) = Refl′ (s≤s l≤l′)
  s≤s//′ (Step′ para) = Step′ (s≤s//′ para)

  s≤s// : {l l′ m m′ : ℕ} → {l≤l′ : l ≤ l′} → {m≤m′ : m ≤ m′} →
          l≤l′ // m≤m′ → s≤s l≤l′ // s≤s m≤m′
  s≤s// (Refl l≤l′) = Refl (s≤s l≤l′)
  s≤s// (Step para) = Step (s≤s// para)

  //′-step : {l l′ m m′ : ℕ} → {m≤l : m ≤ l} → {m′≤l′ : m′ ≤ l′} →
          m≤l //′ m′≤l′ → ≤-step m≤l //′ ≤-step m′≤l′
  //′-step (Refl′ m≤l) = Refl′ (≤-step m≤l)
  //′-step (Step′ para′) = Step′ (//′-step para′)

  //-step : {l l′ m m′ : ℕ} → {l≤l′ : l ≤ l′} → {m≤m′ : m ≤ m′} →
          l≤l′ // m≤m′ → ≤-step l≤l′ // ≤-step m≤m′
  //-step (Refl l≤l′) = Refl (≤-step l≤l′)
  //-step (Step para) = Step (//-step para)

  //′-refl : {l m : ℕ} →
          (l≤l : l ≤ l) → (m≤m : m ≤ m) → l ≤′ m →
          l≤l //′ m≤m
  //′-refl {zero} z≤n z≤n ≤′-refl = Refl′ z≤n
  //′-refl {suc l} (s≤s l≤l) (s≤s m≤m) ≤′-refl =
    s≤s//′ (//′-refl l≤l m≤m ≤′-refl)
  //′-refl l≤l (s≤s m≤m) (≤′-step l≤′m) = Step′ (//′-refl l≤l m≤m l≤′m)

  //-refl : {l m : ℕ} → (l≤l : l ≤ l) → (m≤m : m ≤ m) →
          m ≤′ l → l≤l // m≤m
  //-refl {zero} z≤n z≤n ≤′-refl = Refl z≤n
  //-refl {suc l} (s≤s l≤l) (s≤s m≤m) ≤′-refl =
    s≤s// (//-refl l≤l m≤m ≤′-refl)
  //-refl (s≤s l≤l) m≤m (≤′-step para) = Step (//-refl l≤l m≤m para)

//→//′ : {l l′ m m′ : ℕ} → {l≤l′ : l ≤ l′} → {m≤m′ : m ≤ m′} →
        (para : l≤l′ // m≤m′) → 
        Σ[ m≤l ∈ m ≤ l ] Σ[ m′≤l′ ∈ m′ ≤ l′ ] m≤l //′ m′≤l′ 
//→//′ {l} {l′} (Refl l≤l′) with Data.Nat.compare l l′
//→//′ (Refl l≤l′) | less m k =
  (n≤m+n 0 m , n≤m+n 0 (suc (m + k)) ,
   //′-refl (n≤m+n 0 m) (n≤m+n 0 (suc (m + k))) (≤⇒≤′ l≤l′))
//→//′ (Refl l≤l′) | equal m = (n≤m+n 0 m , n≤m+n 0 m , Refl′ (n≤m+n 0 m))
//→//′ (Refl l≤l′) | greater m k = ⊥-elim (1+n≰n sucm≤m)
  where sucm≤m : suc m ≤ m
        sucm≤m = ≤-trans (m≤m+n (suc m) k) l≤l′
//→//′ (Step para) with //→//′ para
... | (m≤l , m′≤l′ , m≤l//′m′≤l′ ) =
  (≤-step m≤l , ≤-step m′≤l′ , //′-step m≤l//′m′≤l′)

//′→// : {l m l′ m′ : ℕ} → {m≤l : m ≤ l} → {m′≤l′ : m′ ≤ l′} →
        (para′ : m≤l //′ m′≤l′) → 
        Σ[ l≤l′ ∈ l ≤ l′ ] Σ[ m≤m′ ∈ m ≤ m′ ] l≤l′ // m≤m′ 
//′→// {l} {m} (Refl′ m≤l) with Data.Nat.compare m l
//′→// (Refl′ m≤l) | less m k =
  (n≤m+n 0 (suc (m + k)) , n≤m+n 0 m ,
   //-refl (n≤m+n 0 (suc (m + k))) (n≤m+n 0 m) (≤⇒≤′ m≤l))
//′→// (Refl′ m≤l) | equal m = (n≤m+n 0 m , n≤m+n 0 m , Refl (n≤m+n 0 m))
//′→// (Refl′ m≤l) | greater m k = ⊥-elim (1+n≰n sucm≤m)
  where sucm≤m : suc m ≤ m
        sucm≤m = ≤-trans (m≤m+n (suc m) k) m≤l
//′→// (Step′ para′) with //′→// para′
... | (l≤l′ , m≤m′ , l≤l′//m≤m′) =
  (≤-step l≤l′ , ≤-step m≤m′ , //-step l≤l′//m≤m′)

//-leq : {l l' m m' : ℕ} → {l≤l' : l ≤ l'} → {m≤m' : m ≤ m'} →
        l≤l' // m≤m' → m ≤′ l
//-leq {m} (Refl l≤l') = ≤′-refl
//-leq (Step para) with //-leq para
... | m≤l = ≤′-step m≤l

data parallel-view₂ : 
    {l l₁ l₂ m m₁ m₂ : ℕ} → 
    {l≤l₁ : l ≤ l₁} → {m≤m₁ : m ≤ m₁} →
    {l≤l₂ : l ≤ l₂} → {m≤m₂ : m ≤ m₂} →
    l≤l₁ // m≤m₁ → l≤l₂ // m≤m₂ → Set where
  Refl : {l l₁ l₂ : ℕ} → 
         {l≤l₁ : l ≤ l₁} → {l≤l₂ : l ≤ l₂} →
         parallel-view₂ (Refl l≤l₁) (Refl l≤l₂)
  Step : {l l₁ l₂ m m₁ m₂ : ℕ} → 
         {l≤l₁ : l ≤ l₁} → {m≤m₁ : m ≤ m₁} →
         {l≤l₁//m≤m₁ : l≤l₁ // m≤m₁} →
         {l≤l₂ : l ≤ l₂} → {m≤m₂ : m ≤ m₂} →
         {l≤l₂//m≤m₂ : l≤l₂ // m≤m₂} →
         parallel-view₂ l≤l₁//m≤m₁ l≤l₂//m≤m₂ →
         parallel-view₂ (Step l≤l₁//m≤m₁) (Step l≤l₂//m≤m₂)

parallel-same₂ : {l l₁ l₂ m m₁ m₂ : ℕ} → 
        {l≤l₁ : l ≤ l₁} → {m≤m₁ : m ≤ m₁} →
        (l≤l₁//m≤m₁ : l≤l₁ // m≤m₁) →
        {l≤l₂ : l ≤ l₂} → {m≤m₂ : m ≤ m₂} →
        (l≤l₂//m≤m₂ : l≤l₂ // m≤m₂) →
        parallel-view₂ l≤l₁//m≤m₁ l≤l₂//m≤m₂
parallel-same₂ (Refl l≤l₁) (Refl l≤l₂) = Refl
parallel-same₂ (Refl l≤l₁) (Step l≤l₂//m≤m₂)
  with //→//′ l≤l₂//m≤m₂
... | (1+l≤l , _ , _) = ⊥-elim (1+n≰n 1+l≤l)
parallel-same₂ (Step l≤l₁//m≤m₁) (Refl l≤l₂)
  with //→//′ l≤l₁//m≤m₁
... | (1+l≤l , _ , _) = ⊥-elim (1+n≰n 1+l≤l)
parallel-same₂ (Step l≤l₁//m≤m₁) (Step l≤l₂//m≤m₂) =
  Step (parallel-same₂ l≤l₁//m≤m₁ l≤l₂//m≤m₂)

data parallel-view₃ : 
    {l l₁ l₂ m m₁ m₂ : ℕ} → 
    {l≤l₁ : l ≤ l₁} → {m≤m₁ : m ≤ m₁} →
    {l≤l₂ : l ≤ l₂} → {m≤m₂ : m ≤ m₂} →
    {l₁≤l₂ : l₁ ≤ l₂} → {m₁≤m₂ : m₁ ≤ m₂} →
    l≤l₁ // m≤m₁ → l₁≤l₂ // m₁≤m₂ → l≤l₂ // m≤m₂ → Set where
  Refl : {l l₁ l₂ : ℕ} → 
         {l≤l₁ : l ≤ l₁} → {l₁≤l₂ : l₁ ≤ l₂} → {l≤l₂ : l ≤ l₂} →
         parallel-view₃ (Refl l≤l₁) (Refl l₁≤l₂) (Refl l≤l₂)
  Step : {l l₁ l₂ m m₁ m₂ : ℕ} → 
         {l≤l₁ : l ≤ l₁} → {m≤m₁ : m ≤ m₁} →
         {l≤l₁//m≤m₁ : l≤l₁ // m≤m₁} →
         {l₁≤l₂ : l₁ ≤ l₂} → {m₁≤m₂ : m₁ ≤ m₂} →
         {l₁≤l₂//m₁≤m₂ : l₁≤l₂ // m₁≤m₂} →
         {l≤l₂ : l ≤ l₂} → {m≤m₂ : m ≤ m₂} →
         {l≤l₂//m≤m₂ : l≤l₂ // m≤m₂} →
         parallel-view₃ l≤l₁//m≤m₁ l₁≤l₂//m₁≤m₂ l≤l₂//m≤m₂ →
         parallel-view₃ (Step l≤l₁//m≤m₁) (Step l₁≤l₂//m₁≤m₂)
                        (Step l≤l₂//m≤m₂)

parallel-same₃ : {l l₁ l₂ m m₁ m₂ : ℕ} → 
        {l≤l₁ : l ≤ l₁} → {m≤m₁ : m ≤ m₁} →
        (l≤l₁//m≤m₁ : l≤l₁ // m≤m₁) →
        {l₁≤l₂ : l₁ ≤ l₂} → {m₁≤m₂ : m₁ ≤ m₂} →
        (l₁≤l₂//m₁≤m₂ : l₁≤l₂ // m₁≤m₂) →
        {l≤l₂ : l ≤ l₂} → {m≤m₂ : m ≤ m₂} →
        (l≤l₂//m≤m₂ : l≤l₂ // m≤m₂) →
        parallel-view₃ l≤l₁//m≤m₁ l₁≤l₂//m₁≤m₂ l≤l₂//m≤m₂
parallel-same₃ (Refl l≤l₁) (Refl l₁≤l₂) (Refl l≤l₂) = Refl
parallel-same₃ (Refl l≤l₁) (Refl l₁≤l₂) (Step l≤l₂//m≤m₂)
  with //→//′ l≤l₂//m≤m₂
... | (1+l≤l , _ , _) = ⊥-elim (1+n≰n 1+l≤l)
parallel-same₃ (Refl l≤l₁) (Step l₁≤l₂//m₁≤m₂) (Refl l≤l₂)
  with //→//′ l₁≤l₂//m₁≤m₂
... | (1+l≤l , _ , _) = ⊥-elim (1+n≰n 1+l≤l)
parallel-same₃ (Refl l≤l₁) (Step l₁≤l₂//m₁≤m₂) (Step l≤l₂//m≤m₂)
  with //→//′ l≤l₂//m≤m₂
... | (1+l≤l , _ , _) = ⊥-elim (1+n≰n 1+l≤l)
parallel-same₃ (Step l≤l₁//m≤m₁) (Refl l₁≤l₂) (Refl l≤l₂)
  with //→//′ l≤l₁//m≤m₁
... | (1+l≤l , _ , _) = ⊥-elim (1+n≰n 1+l≤l)
parallel-same₃ (Step l≤l₁//m≤m₁) (Refl l₁≤l₂) (Step l≤l₂//m≤m₂)
  with //→//′ l≤l₁//m≤m₁
... | (_ , 1+l≤l , _) = ⊥-elim (1+n≰n 1+l≤l)
parallel-same₃ (Step l≤l₁//m≤m₁) (Step l₁≤l₂//m₁≤m₂) (Refl l≤l₂)
  with //→//′ l≤l₁//m≤m₁
... | (1+l≤l , _ , _) = ⊥-elim (1+n≰n 1+l≤l)
parallel-same₃ (Step l≤l₁//m≤m₁) (Step l₁≤l₂//m₁≤m₂) (Step l≤l₂//m≤m₂) =
  Step (parallel-same₃ l≤l₁//m≤m₁ l₁≤l₂//m₁≤m₂ l≤l₂//m≤m₂)

liftAList≤ : {F : Code} → {l l′ m m′ : ℕ} →
        {l≤l′ : l ≤ l′} → {m≤m′ : m ≤ m′} → l≤l′ // m≤m′ →
        (σ : AList F l m) → AList F l′ m′
liftAList≤ {l≤l′ = l≤l′} {m≤m′ = .l≤l′} (Refl .l≤l′) σ with AListInvAnil σ
... | σ≡anil = anil
liftAList≤ {l = suc l} {l′ = suc l′} {l≤l′ = s≤s l≤l′} {m≤m′ = m≤m′} (Step para) σ
  with //→//′ para
... | (m≤l , m′≤l′ , m≤l//′m′≤l)
  with AListInvAsnoc′ m≤l σ
... | (σ′ , t , x , eq)
  = liftAList≤ para σ′ asnoc lift≤ l≤l′ t / inject≤ x (s≤s l≤l′)

private
  liftAList≤-append : {F : Code} → {l l′ l′′ m m′ m′′ : ℕ} →
          {l≤l′ : l ≤ l′} → {m≤m′ : m ≤ m′} →
          (l≤l′//m≤m′ : l≤l′ // m≤m′) →
          {l′≤l′′ : l′ ≤ l′′} → {m′≤m′′ : m′ ≤ m′′} →
          (l′≤l′′//m′≤m′′ : l′≤l′′ // m′≤m′′) →
          {l≤l′′ : l ≤ l′′} → {m≤m′′ : m ≤ m′′} →
          (l≤l′′//m≤m′′ : l≤l′′ // m≤m′′) →
          (σ : AList F l m) →
          liftAList≤ l′≤l′′//m′≤m′′ (liftAList≤ l≤l′//m≤m′ σ) ≡
          liftAList≤ l≤l′′//m≤m′′ σ
  liftAList≤-append {l≤l′} {m≤m′} l≤l′//m≤m′ {l′≤l′′} {m′≤m′′} l′≤l′′//m′≤m′′ {l≤l′′} {m≤m′′} l≤l′′//m≤m′′ σ
    with parallel-same₃ l≤l′//m≤m′ l′≤l′′//m′≤m′′ l≤l′′//m≤m′′ 
  liftAList≤-append {l≤l′ = l≤l′} {m≤m′ = .l≤l′} .(Refl l≤l′) {l′≤l′′} {.l′≤l′′} .(Refl l′≤l′′) {l≤l′′} {.l≤l′′} .(Refl l≤l′′) σ | Refl
    with AListInvAnil σ
  ... | σ≡anil rewrite σ≡anil = refl
  liftAList≤-append {F} {.(suc l)} {.(suc l₁)} {.(suc l₂)} {.(suc l)} {m′} {m′′} {.(s≤s {l} {l₁} l≤l₁)} {m≤m′} .(Step {l} {l₁} {suc l} {m′} {l≤l₁} {m≤m′} l≤l₁//m≤m₁) {.(s≤s {l₁} {l₂} l₁≤l₂)} {m′≤m′′} .(Step {l₁} {l₂} {m′} {m′′} {l₁≤l₂} {m′≤m′′} l₁≤l₂//m₁≤m₂) {.(s≤s {l} {l₂} l≤l₂)} {m≤m′′} .(Step {l} {l₂} {suc l} {m′′} {l≤l₂} {m≤m′′} l≤l₂//m≤m₂) (anil {m = .(suc l)}) | Step {l = l} {l₁ = l₁} {l₂ = l₂} {m = .(suc l)} {m₁ = .m′} {m₂ = .m′′} {l≤l₁ = l≤l₁} {m≤m₁ = .m≤m′} {l≤l₁//m≤m₁ = l≤l₁//m≤m₁} {l₁≤l₂ = l₁≤l₂} {m₁≤m₂ = .m′≤m′′} {l₁≤l₂//m₁≤m₂ = l₁≤l₂//m₁≤m₂} {l≤l₂ = l≤l₂} {m≤m₂ = .m≤m′′} {l≤l₂//m≤m₂ = l≤l₂//m≤m₂} view₃
    with //→//′ l≤l₁//m≤m₁
  ... | (1+l≤l , _ , _) = ⊥-elim (1+n≰n 1+l≤l)
  liftAList≤-append {F} {.(suc l)} {.(suc l₁)} {.(suc l₂)} {m} {m′} {m′′} {.(s≤s {l} {l₁} l≤l₁)} {m≤m′} .(Step {l} {l₁} {m} {m′} {l≤l₁} {m≤m′} l≤l₁//m≤m₁) {.(s≤s {l₁} {l₂} l₁≤l₂)} {m′≤m′′} .(Step {l₁} {l₂} {m′} {m′′} {l₁≤l₂} {m′≤m′′} l₁≤l₂//m₁≤m₂) {.(s≤s {l} {l₂} l≤l₂)} {m≤m′′} .(Step {l} {l₂} {m} {m′′} {l≤l₂} {m≤m′′} l≤l₂//m≤m₂) (_asnoc_/_ {l = l} {m = .m} σ t x) | Step {l = .l} {l₁ = l₁} {l₂ = l₂} {m = .m} {m₁ = .m′} {m₂ = .m′′} {l≤l₁ = l≤l₁} {m≤m₁ = .m≤m′} {l≤l₁//m≤m₁ = l≤l₁//m≤m₁} {l₁≤l₂ = l₁≤l₂} {m₁≤m₂ = .m′≤m′′} {l₁≤l₂//m₁≤m₂ = l₁≤l₂//m₁≤m₂} {l≤l₂ = l≤l₂} {m≤m₂ = .m≤m′′} {l≤l₂//m≤m₂ = l≤l₂//m≤m₂} view₃
    with //→//′ l≤l₁//m≤m₁ | //→//′ l₁≤l₂//m₁≤m₂ | //→//′ l≤l₂//m≤m₂
  ... | (m≤l , m′′≤l₂ , _) | (m′≤l₁ , m′′≤₂l₂ , _) | (m≤₂l , m′≤₂l₁ , _)
    rewrite liftAList≤-append l≤l₁//m≤m₁ l₁≤l₂//m₁≤m₂ l≤l₂//m≤m₂ σ
          | fuse-lift l≤l₁ l₁≤l₂ l≤l₂ t
          | inject≤-trans x (s≤s l≤l₁) (s≤s l₁≤l₂) (s≤s l≤l₂) 
    = refl

≤→// : {m m' l : ℕ} → (m≤m' : m ≤ m') → (m≤′l : m ≤′ l) →
        Σ[ l' ∈ ℕ ] Σ[ l≤l' ∈ l ≤ l' ] l≤l' // m≤m'
≤→// {m} {m'} m≤m' ≤′-refl = (m' , m≤m' , Refl m≤m')
≤→// {m} {m'} {suc l} m≤m' (≤′-step m≤′l) with ≤→// m≤m' m≤′l
... | (l' , l≤l' , para) = (suc l' , s≤s l≤l' , Step para)

σ→// : {F : Code} → {m m' l : ℕ} → (σ : AList F l m) → (m≤m' : m ≤ m') →
        Σ[ l' ∈ ℕ ] Σ[ l≤l' ∈ l ≤ l' ] l≤l' // m≤m'
σ→// σ m≤m' = ≤→// m≤m' (AListLeq σ)

-- append two substitutions after lifting the second one
_+⟨_⟩_ : {F : Code} → {l m l′ m′ n : ℕ} →
        {l≤l′ : l ≤ l′} → {m≤m′ : m ≤ m′} →
        (ρ : AList F m′ n) → l≤l′ // m≤m′ → (σ : AList F l m) →
        AList F l′ n
ρ +⟨ para ⟩ σ = ρ ++ liftAList≤ para σ

-- sub σ : transform σ to a function from Fin m to μ F m'
sub : {F : Code} → {m m′ : ℕ} →
      (σ : AList F m m′) → Fin m → μ F m′
sub anil = ⟪_⟫
sub (σ asnoc t / x) = ⟪cata⟫ (sub σ) ∘ (t for x)

-- applyUnifier f t : apply unifier f to t
applyUnifier : {F : Code} → {m m′ : ℕ} →
        (f : Fin m → μ F m′) → (t : μ F m) → μ F m′
applyUnifier f t = cata ⟨_⟩ f t

-- applyUnifier≤ f t : apply unifier f to t after lifting
applyUnifier≤ : {F : Code} → {m m′ m′′ : ℕ} →
        (f : Fin m′′ → μ F m′) →
        (m≤m′′ : m ≤ m′′) → (t : μ F m) → μ F m′
applyUnifier≤ {F} f m≤m′′ t = cata ⟨_⟩ f (lift≤ m≤m′′ t)

-- applySub σ t : apply substitution σ to t
applySub : {F : Code} → {m m′ : ℕ} →
           (σ : AList F m m′) → (t : μ F m) → μ F m′
applySub σ t = ⟪cata⟫ (sub σ) t

applySub-f : {F : Code} → (G : Code) → {m m′ : ℕ} →
           (σ : AList F m m′) → (d : ⟦ G ⟧ (μ F m)) → ⟦ G ⟧ (μ F m′)
applySub-f G σ d = ⟪fmap⟫ G (sub σ) d

applySub-f≤ : {F : Code} → (G : Code) → {m m′ m′′ : ℕ} →
           (σ : AList F m′′ m′) → (m≤m′′ : m ≤ m′′) → (d : ⟦ G ⟧ (μ F m)) → ⟦ G ⟧ (μ F m′)
applySub-f≤ G σ m≤m′′ d = ⟪fmap⟫≤ G (sub σ) m≤m′′ d

-- applySub σ m≤m′ t : apply substitution σ to t after lifting
applySub≤ : {F : Code} → {m m′ m′′ : ℕ} →
        (σ : AList F m′′ m′) → (m≤m′′ : m ≤ m′′) →
        (t : μ F m) → μ F m′
applySub≤ σ m≤m′′ t = applySub σ (lift≤ m≤m′′ t)

private
  ⟨liftAList≤-same⟩ : {F : Code} → (G : Code) → {l m : ℕ} →
          {l≤l : l ≤ l} → {m≤m : m ≤ m} →
          (l≤l//m≤m : l≤l // m≤m) → 
          (σ : AList F l m) →
          (d : ⟦ G ⟧ (μ F l)) →
          (r : ⟦ G ⟧′ (λ t → applySub (liftAList≤ l≤l//m≤m σ) t ≡
                              applyUnifier (sub σ) t) d) →
          fmap G ⟨_⟩ (sub (liftAList≤ l≤l//m≤m σ)) d ≡
          fmap G ⟨_⟩ (sub σ) d
  ⟨liftAList≤-same⟩ U l≤l//m≤m σ tt tt = refl
  ⟨liftAList≤-same⟩ I l≤l//m≤m σ d r = r
  ⟨liftAList≤-same⟩ (G₁ ⊕ G₂) l≤l//m≤m σ (inj₁ d) r =
    cong inj₁ (⟨liftAList≤-same⟩ G₁ l≤l//m≤m σ d r)
  ⟨liftAList≤-same⟩ (G₁ ⊕ G₂) l≤l//m≤m σ (inj₂ d) r =
    cong inj₂ (⟨liftAList≤-same⟩ G₂ l≤l//m≤m σ d r)
  ⟨liftAList≤-same⟩ (G₁ ⊗ G₂) l≤l//m≤m σ (d₁ , d₂) (r₁ , r₂) =
    cong₂ _,_ (⟨liftAList≤-same⟩ G₁ l≤l//m≤m σ d₁ r₁)
             (⟨liftAList≤-same⟩ G₂ l≤l//m≤m σ d₂ r₂)

  ⟪liftAList≤-same⟫ : {F : Code} → {l m : ℕ} →
          {l≤l : l ≤ l} → {m≤m : m ≤ m} →
          (para : l≤l // m≤m) → 
          (σ : AList F l m) →
          (x : Fin l) →
          sub (liftAList≤ para σ) x ≡ sub σ x
  ⟪liftAList≤-same⟫ {l≤l = l≤l} {m≤m = .l≤l} (Refl .l≤l) σ x
    with AListInvAnil σ
  ... | σ≡anil rewrite σ≡anil = refl
  ⟪liftAList≤-same⟫ {F} {._} {m≤m} (Step {l≤l′ = l≤l′} para) σ x
    with //→//′ (Step para)
  ... | (m≤1+l , m′≤1+l′ , m≤1+l//′m′≤1+l′)
    with //→//′ para
  ... | (m≤l , m′≤l′ , m≤l//′m′≤l′)
    with AListInvAsnoc′ m≤l σ
  ... | (σ′ , t , y , eq) rewrite inject≤-id y (s≤s l≤l′) | eq
    with thick y x
  ... | nothing
    rewrite cata-id (λ x → ⟪ inject≤ x l≤l′ ⟫)
                    (λ x → cong ⟪_⟫ (inject≤-id x l≤l′)) t
          | cata-ext (sub (liftAList≤ para σ′)) (sub σ′)
                     (⟪liftAList≤-same⟫ para σ′) t
    = refl
  ... | just x′ = ⟪liftAList≤-same⟫ para σ′ x′

  liftAList≤-same : {F : Code} → {l m : ℕ} →
          {l≤l : l ≤ l} → {m≤m : m ≤ m} →
          (l≤l//m≤m : l≤l // m≤m) → 
          (σ : AList F l m) →
          (t : μ F l) →
          applySub (liftAList≤ l≤l//m≤m σ) t ≡
          applyUnifier (sub σ) t
  liftAList≤-same {F} {l≤l = l≤l} {m≤m = m≤m} l≤l//m≤m σ t =
    ind (λ t → applySub (liftAList≤ l≤l//m≤m σ) t ≡
                   applyUnifier (sub σ) t)
        (λ d r → cong ⟨_⟩ (⟨liftAList≤-same⟩ F l≤l//m≤m σ d r))
        (λ x → ⟪liftAList≤-same⟫ l≤l//m≤m σ x)
        t

private
  liftAList-max : {F : Code} → {l l′ m m′ : ℕ} →
          (l≤l′ : l ≤ l′) → {l≤1+l′ : l ≤ suc l′} →
          (m≤m′ : m ≤ m′) → {m≤1+m′ : m ≤ suc m′} →
          (para : l≤1+l′ // m≤1+m′) →
          (σ : AList F l m) →
          sub (liftAList≤ para σ) (fromℕ l′) ≡ ⟪ fromℕ m′ ⟫
  liftAList-max l≤l′ {l≤1+l′} m≤m′ {.l≤1+l′} (Refl .l≤1+l′) σ = refl
  liftAList-max {F} {.(suc l)} {.(suc l′)} {m} {m′} (s≤s {m = l} {n = l′} l≤l′) {.(s≤s {l} {suc l′} l≤1+l′)} m≤m′ {m≤1+m′} (Step {l = .l} {l′ = .(suc l′)} {m = .m} {m′ = .(suc m′)} {l≤l′ = l≤1+l′} {m≤m′ = .m≤1+m′} para) σ
    with //→//′ (Step para)
  ... | (m≤1+l , m′≤1+l′ , m≤1+l//′m′≤1+l′)
    with //→//′ para
  ... | (m≤l , 1+m′≤1+l′ , m≤l//′1+m′≤1+l′ )
    with AListInvAsnoc′ m≤l σ
  ... | (σ′ , t , y , eq)
    rewrite thick-max y l≤l′ l≤1+l′
    = liftAList-max l≤l′ m≤m′ para σ′

liftFun′ : {F : Code} → {l m l′ m′ : ℕ} →
        {m≤l : m ≤ l} → {m′≤l′ : m′ ≤ l′} →
        (para′ : m≤l //′ m′≤l′) → 
        (f : Fin l → μ F m) → (x : Fin l′) → μ F m′
liftFun′ (Refl′ m≤l) f x = f x
liftFun′ {l = l} (Step′ m≤l) f x with finCompare x
liftFun′ (Step′ {l′ = l′} {m′ = m′} m≤l//′m′≤l′) f .(fromℕ l′) | max =
  ⟪ fromℕ m′ ⟫
liftFun′ (Step′ {l′ = l′} {m′ = m′} m≤l//′m′≤l′) f .(inject≤ x (n≤1+n l′)) | less x =
  lift≤ (n≤1+n m′) (liftFun′ m≤l//′m′≤l′ f x)

liftFun : {F : Code} → {l m l′ m′ : ℕ} →
        {l≤l′ : l ≤ l′} → {m≤m′ : m ≤ m′} →
        (para : l≤l′ // m≤m′) → 
        (f : Fin l → μ F m) → (x : Fin l′) → μ F m′
liftFun para f x with //→//′ para
... | (m≤l , m′≤l′ , m≤l//′m′≤l′) = liftFun′ m≤l//′m′≤l′ f x

private
  liftFun-liftFun′ : {F : Code} → {l l′ m m′ : ℕ} →
          (l≤l′ : l ≤ l′) → (m≤m′ : m ≤ m′) →
          (m≤l : m ≤ l) → (m′≤l′ : m′ ≤ l′) →
          (l≤l′//m≤m′ : l≤l′ // m≤m′) →
          (m≤l//′m′≤l′ : m≤l //′ m′≤l′) →
          //→//′ l≤l′//m≤m′ ≡ (m≤l , m′≤l′ , m≤l//′m′≤l′) →
          (σ : AList F l m) →
          (x : Fin l′) →
          liftFun l≤l′//m≤m′ (sub σ) x ≡
          liftFun′ m≤l//′m′≤l′ (sub σ) x
  liftFun-liftFun′ l≤l′ m≤m′ m≤l m′≤l′ l≤l′//m≤m′ m≤l//′m′≤l′ eq σ x
    rewrite eq = refl

mutual
  lift-sub-commute1 : {F : Code} → {l l′ m m′ : ℕ} →
          {l≤l′ : l ≤ l′} → {m≤m′ : m ≤ m′} →
          (l≤l′//m≤m′ : l≤l′ // m≤m′) →
          {l≤1+l′ : l ≤ suc l′} → {m≤1+m′ : m ≤ suc m′} →
          (l≤1+l′//m≤1+m′ : l≤1+l′ // m≤1+m′) →
          (σ : AList F l m) → (x : Fin l′) →
          sub (liftAList≤ l≤1+l′//m≤1+m′ σ)
              (inject≤ x (n≤1+n l′)) ≡
          lift≤ (n≤1+n m′)
                (sub (liftAList≤ l≤l′//m≤m′ σ) x)
  lift-sub-commute1 {l≤l′ = l≤l′} {m≤m′ = m≤m′} l≤l′//m≤m′ l≤1+l′//m≤1+m′ σ x
    with parallel-same₂ l≤l′//m≤m′ l≤1+l′//m≤1+m′ 
  lift-sub-commute1 {l≤l′ = l≤l′} {m≤m′ = .l≤l′} .(Refl l≤l′) {l≤1+l′} {.l≤1+l′} .(Refl l≤1+l′) σ x | Refl = refl
  lift-sub-commute1 {F} {.(suc l)} {.(suc l₁)} {.(suc l)} {m′} {.(s≤s {l} {l₁} l≤l₁)} {m≤m′} .(Step {l} {l₁} {suc l} {m′} {l≤l₁} {m≤m′} l≤l₁//m≤m₁) {.(s≤s {l} {suc l₁} l≤l₂)} {m≤1+m′} .(Step {l} {suc l₁} {suc l} {suc m′} {l≤l₂} {m≤1+m′} l≤l₂//m≤m₂) (anil {m = .(suc l)}) x | Step {l = l} {l₁ = l₁} {l₂ = .(suc l₁)} {m = .(suc l)} {m₁ = .m′} {m₂ = .(suc m′)} {l≤l₁ = l≤l₁} {m≤m₁ = .m≤m′} {l≤l₁//m≤m₁ = l≤l₁//m≤m₁} {l≤l₂ = l≤l₂} {m≤m₂ = .m≤1+m′} {l≤l₂//m≤m₂ = l≤l₂//m≤m₂} pview
    with //→//′ l≤l₂//m≤m₂
  ... | (1+l≤l , _ , _) = ⊥-elim (1+n≰n 1+l≤l)
  lift-sub-commute1 {F} {.(suc l)} {.(suc l₁)} {m} {m₁} {.(s≤s {l} {l₁} l≤l₁)} {m≤m₁} .(Step {l} {l₁} {m} {m₁} {l≤l₁} {m≤m₁} l≤l₁//m≤m₁) {.(s≤s {l} {suc l₁} l≤1+l₁)} {m≤1+m₁} .(Step {l} {suc l₁} {m} {suc m₁} {l≤1+l₁} {m≤1+m₁} l≤1+l₁//m≤1+m₁) (_asnoc_/_ {l = l} {m = .m} σ t y) x | Step {l = .l} {l₁ = l₁} {l₂ = .(suc l₁)} {m = .m} {m₁ = .m₁} {m₂ = .(suc m₁)} {l≤l₁ = l≤l₁} {m≤m₁ = .m≤m₁} {l≤l₁//m≤m₁ = l≤l₁//m≤m₁} {l≤l₂ = l≤1+l₁} {m≤m₂ = .m≤1+m₁} {l≤l₂//m≤m₂ = l≤1+l₁//m≤1+m₁} pview
    with thick (inject≤ y (s≤s l≤l₁)) x
       | inspect (thick (inject≤ y (s≤s l≤l₁))) x
  ... | nothing | [ eq ] rewrite thick-inject-nothing y x l≤l₁ l≤1+l₁ eq
    = begin
        cata ⟨_⟩ (sub (liftAList≤ l≤1+l₁//m≤1+m₁ σ))
             (lift≤ l≤1+l₁ t)
      ≡⟨ cong
            (cata ⟨_⟩ (sub (liftAList≤ l≤1+l₁//m≤1+m₁ σ))) (sym (fuse-lift l≤l₁ (n≤1+n l₁) l≤1+l₁ t)) ⟩
        cata ⟨_⟩ (sub (liftAList≤ l≤1+l₁//m≤1+m₁ σ))
               (lift≤ (n≤1+n l₁) (lift≤ l≤l₁ t))
      ≡⟨ refl ⟩
        cata ⟨_⟩ (sub (liftAList≤ l≤1+l₁//m≤1+m₁ σ))
               (cata ⟨_⟩ (λ x → ⟪ inject≤ x (n≤1+n l₁) ⟫)
             (lift≤ l≤l₁ t))
      ≡⟨ fuse (sub (liftAList≤ l≤1+l₁//m≤1+m₁ σ))
              (λ x → ⟪ inject≤ x (n≤1+n l₁) ⟫) (lift≤ l≤l₁ t) ⟩
        cata ⟨_⟩
             (cata {F} ⟨_⟩ (sub (liftAList≤ l≤1+l₁//m≤1+m₁ σ))
              ∘ (λ x → ⟪ inject≤ x (n≤1+n l₁) ⟫))
             (lift≤ l≤l₁ t)
      ≡⟨ cata-ext
          ((cata {F} ⟨_⟩ (sub (liftAList≤ l≤1+l₁//m≤1+m₁ σ))
           ∘ (λ x → ⟪ inject≤ x (n≤1+n l₁) ⟫)))
          ((cata ⟨_⟩ (λ x → ⟪ inject≤ x (n≤1+n m₁) ⟫)
           ∘ (sub (liftAList≤ l≤l₁//m≤m₁ σ))))
          (lift-sub-commute1 l≤l₁//m≤m₁ l≤1+l₁//m≤1+m₁  σ)
          (lift≤ l≤l₁ t) ⟩
        cata ⟨_⟩ 
             (cata ⟨_⟩ (λ x → ⟪ inject≤ x (n≤1+n m₁) ⟫)
              ∘ (sub (liftAList≤ l≤l₁//m≤m₁ σ)))
             (lift≤ l≤l₁ t)
      ≡⟨ sym (fuse (λ x → ⟪ inject≤ x (n≤1+n m₁) ⟫)
                   (sub (liftAList≤ l≤l₁//m≤m₁ σ))
                   (lift≤ l≤l₁ t)) ⟩
        cata ⟨_⟩ (λ x → ⟪ inject≤ x (n≤1+n m₁) ⟫)
              (cata ⟨_⟩ (sub (liftAList≤ l≤l₁//m≤m₁ σ))
                         (lift≤ l≤l₁ t))
      ≡⟨ refl ⟩
        lift≤ (n≤1+n m₁)
              (cata ⟨_⟩ (sub (liftAList≤ l≤l₁//m≤m₁ σ))
                         (lift≤ l≤l₁ t))
      ∎
  ... | just x′ | [ eq ] rewrite thick-inject-just y x x′ l≤l₁ l≤1+l₁ eq
    = lift-sub-commute1 l≤l₁//m≤m₁ l≤1+l₁//m≤1+m₁ σ x′
  
  ⟪lift-sub-commute⟫ : {F : Code} → {l l′ m m′ : ℕ} →
          (l≤l′ : l ≤ l′) → (m≤m′ : m ≤ m′) →
          (m≤l : m ≤ l) → (m′≤l′ : m′ ≤ l′) →
          (l≤l′//m≤m′ : l≤l′ // m≤m′) →
          (m≤l//′m′≤l′ : m≤l //′ m′≤l′) →
          (σ : AList F l m) → (x : Fin l′) →
          sub (liftAList≤ l≤l′//m≤m′ σ) x ≡
          liftFun′ m≤l//′m′≤l′ (sub σ) x
  ⟪lift-sub-commute⟫ l≤l′ m≤m′ m≤l .m≤l l≤l′//m≤m′ (Refl′ .m≤l) σ x =
    ⟪liftAList≤-same⟫ l≤l′//m≤m′ σ x
  ⟪lift-sub-commute⟫ l≤l′ m≤m′ m≤l ._ l≤l′//m≤m′ (Step′ m≤l//′m′≤l′) σ x
    with finCompare x
  ⟪lift-sub-commute⟫ l≤1+l′ m≤1+m′ m≤l ._ l≤1+l′//m≤1+m′ (Step′ m≤l//′m′≤l′) σ ._ | max
    with //′→// m≤l//′m′≤l′
  ... | (l≤l′ , m≤m′ , l≤l′//m≤m′)
    = liftAList-max l≤l′ m≤m′ l≤1+l′//m≤1+m′ σ
  ⟪lift-sub-commute⟫ {F} l≤1+l′ m≤1+m′ m≤l ._ l≤1+l′//m≤1+m′ (Step′ {l′ = l′} {m′ = m′} {m′≤l′ = m′≤l′} m≤l//′m′≤l′) σ ._ | less x
    with //′→// m≤l//′m′≤l′
  ... | (l≤l′ , m≤m′ , l≤l′//m≤m′)
    with ⟪lift-sub-commute⟫ l≤l′ m≤m′ m≤l m′≤l′ l≤l′//m≤m′ m≤l//′m′≤l′ σ x
  ... | eq =
    begin
      sub (liftAList≤ l≤1+l′//m≤1+m′ σ)
          (inject≤ x (n≤1+n l′))
    ≡⟨ lift-sub-commute1 l≤l′//m≤m′ l≤1+l′//m≤1+m′ σ x ⟩
      lift≤ (n≤1+n m′)
            (sub (liftAList≤ l≤l′//m≤m′ σ) x)
    ≡⟨ cong (lift≤ (n≤1+n m′)) eq ⟩
      lift≤ (n≤1+n m′) (liftFun′ m≤l//′m′≤l′ (sub σ) x)
    ∎

  lift-sub-commute′ : {F : Code} → {l l′ m m′ : ℕ} →
         {l≤l′ : l ≤ l′} → {m≤m′ : m ≤ m′} →
         {m≤l : m ≤ l} → {m′≤l′ : m′ ≤ l′} →
         (l≤l′//m≤m′ : l≤l′ // m≤m′) →
         (m≤l//′m′≤l′ : m≤l //′ m′≤l′) →
         (σ : AList F l m) → (t : μ F l′) →
         applySub (liftAList≤ l≤l′//m≤m′ σ) t ≡
         applyUnifier (liftFun′ m≤l//′m′≤l′ (sub σ)) t
  lift-sub-commute′ {F} {l≤l′ = l≤l′} {m≤m′ = m≤m′} {m≤l = m≤l} {m′≤l′ = m′≤l′} l≤l′//m≤m′ m≤l//′m′≤l′ σ t =
    begin
      applySub (liftAList≤ l≤l′//m≤m′ σ) t
    ≡⟨ refl ⟩
      applyUnifier (sub (liftAList≤ l≤l′//m≤m′ σ)) t
    ≡⟨ cata-ext (sub (liftAList≤ l≤l′//m≤m′ σ))
                  (liftFun′ m≤l//′m′≤l′ (sub σ))
                  (⟪lift-sub-commute⟫ l≤l′ m≤m′ m≤l m′≤l′
                     l≤l′//m≤m′ m≤l//′m′≤l′ σ) t ⟩
      applyUnifier (liftFun′ m≤l//′m′≤l′ (sub σ)) t
    ∎

lift-sub-commute-f′ : {F : Code} → (G : Code) → {l l′ m m′ : ℕ} →
       {l≤l′ : l ≤ l′} → {m≤m′ : m ≤ m′} →
       {m≤l : m ≤ l} → {m′≤l′ : m′ ≤ l′} →
       (l≤l′//m≤m′ : l≤l′ // m≤m′) →
       (m≤l//′m′≤l′ : m≤l //′ m′≤l′) →
       (σ : AList F l m) → (d : ⟦ G ⟧ (μ F l′)) →
       applySub-f G (liftAList≤ l≤l′//m≤m′ σ) d ≡
       ⟪fmap⟫ G (liftFun′ m≤l//′m′≤l′ (sub σ)) d
lift-sub-commute-f′ {F} G {l≤l′ = l≤l′} {m≤m′ = m≤m′} {m≤l = m≤l} {m′≤l′ = m′≤l′} l≤l′//m≤m′ m≤l//′m′≤l′ σ d =
  begin 
    applySub-f G (liftAList≤ l≤l′//m≤m′ σ) d
  ≡⟨ cata-ext-fmap G (sub (liftAList≤ l≤l′//m≤m′ σ)) (liftFun′ m≤l//′m′≤l′ (sub σ)) (⟪lift-sub-commute⟫ l≤l′ m≤m′ m≤l m′≤l′ l≤l′//m≤m′ m≤l//′m′≤l′ σ) d ⟩
    ⟪fmap⟫ G (liftFun′ m≤l//′m′≤l′ (sub σ)) d
  ∎

-- append two substitutions after lifting the second one
_+⟨_⟩′_ : {F : Code} → {l m l′ m′ n : ℕ} →
        {l≤l′ : l ≤ l′} → {m≤m′ : m ≤ m′} →
        (f₂ : Fin m′ → μ F n) → l≤l′ // m≤m′ → (f₁ : Fin l → μ F m) →
        Fin l′ → μ F n
_+⟨_⟩′_ {l≤l′ = l≤l′} {m≤m′ = m≤m′} f₂ para f₁ x =
  applyUnifier f₂ (liftFun para f₁ x)

f-for : {F : Code} → {m′ m′′ : ℕ} →
        (f : Fin m′′ → μ F m′) → (t : μ F m′) → (x : Fin (suc m′′)) →
        (Fin (suc m′′) → μ F m′)
f-for f t x y with thick₂ x y
... | inj₁ x≡y = t
... | inj₂ (y′ , thinxy′≡y) = f y′

f-for-xx : {F : Code} → {m′ m′′ : ℕ} →
        (f : Fin m′′ → μ F m′) → (t : μ F m′) → (x : Fin (suc m′′)) →
        f-for f t x x ≡ t
f-for-xx f t x rewrite thick₂xx≡inj₁refl x = refl

private
  liftFun-id : {F : Code} → {l l′ : ℕ} →
          {l≤l : l ≤ l} → {l′≤l′ : l′ ≤ l′} →
          (para′ : l′≤l′ //′ l≤l) → (x : Fin l) →
          liftFun′ {F} para′ ⟪_⟫ x ≡ ⟪ x ⟫
  liftFun-id (Refl′ l≤l) x = refl
  liftFun-id (Step′ para′) x with finCompare x
  liftFun-id (Step′ para′) ._ | max = refl
  liftFun-id {F} (Step′ {l′ = l′} {m′≤l′ = m′≤l′} para′) .(inject≤ x (n≤1+n l′)) | less x
    rewrite liftFun-id {F} para′ x = refl

append-nil′ : {F : Code} → {l m l′ : ℕ} →
        (l′≤l : l′ ≤ l) →
        (f : Fin l → μ F m) → (x : Fin l) →
        (f +⟨ Refl l′≤l ⟩′ ⟪_⟫) x ≡ f x
append-nil′ {F} l′≤l f x with //→//′ (Refl l′≤l)
... | (l′≤l′ , l≤l , l′≤l′//′l≤l) rewrite liftFun-id {F} l′≤l′//′l≤l x = refl

mutual
  ⟨sub-++-commute⟩ : (G : Code) → {F : Code} → {m m′ m′′ : ℕ} →
          (σ₂ : AList F m′ m′′) → (σ₁ : AList F m m′) →
          (d : ⟦ G ⟧ (μ F m)) →
          (r : ⟦ G ⟧′ (λ t → cata ⟨_⟩ (cata ⟨_⟩ (sub σ₂) ∘ (sub σ₁)) t
                              ≡ cata ⟨_⟩ (sub (σ₂ ++ σ₁)) t) d) →
          fmap G ⟨_⟩ (cata ⟨_⟩ (sub σ₂) ∘ (sub σ₁)) d
          ≡ fmap G ⟨_⟩ (sub (σ₂ ++ σ₁)) d
  ⟨sub-++-commute⟩ U σ₂ σ₁ tt tt = refl
  ⟨sub-++-commute⟩ I σ₂ σ₁ d r = r
  ⟨sub-++-commute⟩ (G₁ ⊕ G₂) σ₂ σ₁ (inj₁ d) r =
    cong inj₁ (⟨sub-++-commute⟩ G₁ σ₂ σ₁ d r)
  ⟨sub-++-commute⟩ (G₁ ⊕ G₂) σ₂ σ₁ (inj₂ d) r =
    cong inj₂ (⟨sub-++-commute⟩ G₂ σ₂ σ₁ d r)
  ⟨sub-++-commute⟩ (G₁ ⊗ G₂) σ₂ σ₁ (d1 , d2) (r1 , r2) =
    cong₂ _,_ (⟨sub-++-commute⟩ G₁ σ₂ σ₁ d1 r1)
              (⟨sub-++-commute⟩ G₂ σ₂ σ₁ d2 r2)

  ⟪sub-++-commute⟫ : {F : Code} → {m m′ m′′ : ℕ} →
          (σ₂ : AList F m′ m′′) → (σ₁ : AList F m m′) →
          (x : Fin m) →
          cata ⟨_⟩ (sub σ₂) (sub σ₁ x)
          ≡ sub (σ₂ ++ σ₁) x
  ⟪sub-++-commute⟫ σ₂ anil x = refl
  ⟪sub-++-commute⟫ {F} σ₂ (σ₁ asnoc t / y) x =
    begin
      cata ⟨_⟩ (sub σ₂) (cata ⟨_⟩ (sub σ₁) ((t for y) x))
    ≡⟨ fuse (sub σ₂) (sub σ₁) ((t for y) x) ⟩
      cata ⟨_⟩ (cata ⟨_⟩ (sub σ₂) ∘ (sub σ₁)) ((t for y) x)
    ≡⟨ sub-++-commute σ₂ σ₁ ((t for y) x) ⟩
      cata ⟨_⟩ (sub (σ₂ ++ σ₁)) ((t for y) x)
    ≡⟨ refl ⟩
      (cata ⟨_⟩ (sub (σ₂ ++ σ₁)) ∘ (t for y)) x
    ∎

  sub-++-commute : {F : Code} → {m m′ m′′ : ℕ} →
          (σ₂ : AList F m′ m′′) → (σ₁ : AList F m m′) →
          (t : μ F m) →
          cata ⟨_⟩ (cata ⟨_⟩ (sub σ₂) ∘ (sub σ₁)) t
          ≡ cata ⟨_⟩ (sub (σ₂ ++ σ₁)) t
  sub-++-commute {F} σ₂ σ₁ t =
    ind (λ t → cata ⟨_⟩ (cata ⟨_⟩ (sub σ₂) ∘ (sub σ₁)) t
                 ≡ cata ⟨_⟩ (sub (σ₂ ++ σ₁)) t)
        (λ d r → cong ⟨_⟩ (⟨sub-++-commute⟩ F σ₂ σ₁ d r))
        (λ x → ⟪sub-++-commute⟫ σ₂ σ₁ x)
        t

sub-++-commute-fmap : (G : Code) → {F : Code} → {m m′ m′′ : ℕ} →
        (σ₂ : AList F m′ m′′) → (σ₁ : AList F m m′) →
        (d : ⟦ G ⟧ (μ F m)) →
        fmap G ⟨_⟩ (cata ⟨_⟩ (sub σ₂) ∘ (sub σ₁)) d
        ≡ fmap G ⟨_⟩ (sub (σ₂ ++ σ₁)) d
sub-++-commute-fmap U σ₂ σ₁ tt = refl
sub-++-commute-fmap I σ₂ σ₁ d = sub-++-commute σ₂ σ₁ d
sub-++-commute-fmap (G₁ ⊕ G₂) σ₂ σ₁ (inj₁ d) = cong inj₁ (sub-++-commute-fmap G₁ σ₂ σ₁ d)
sub-++-commute-fmap (G₁ ⊕ G₂) σ₂ σ₁ (inj₂ d) = cong inj₂ (sub-++-commute-fmap G₂ σ₂ σ₁ d)
sub-++-commute-fmap (G₁ ⊗ G₂) σ₂ σ₁ (d₁ , d₂) =
  cong₂ _,_ (sub-++-commute-fmap G₁ σ₂ σ₁ d₁) (sub-++-commute-fmap G₂ σ₂ σ₁ d₂)

mutual
    ⟨sub-++-commute⟩′ : (G : Code) → {F : Code} → {m1 m1' m2 m2' m2'' : ℕ} →
            (t : μ F m2'') →
            (σ1 : AList F m1' m1) →
            (m1≤m2' : m1 ≤ m2') →
            (m1'≤m2'' : m1' ≤ m2'') →
            (para : m1'≤m2'' // m1≤m2') →
            (σ2 : AList F m2' m2) →
            (d : ⟦ G ⟧ (μ F m2'')) →
            (r : ⟦ G ⟧′ (λ t → cata ⟨_⟩ (cata ⟨_⟩ (sub σ2) ∘ (sub (liftAList≤ para σ1))) t
                                ≡ cata ⟨_⟩ (sub (σ2 +⟨ para ⟩ σ1)) t) d) →
            fmap G ⟨_⟩ ((cata ⟨_⟩ (sub σ2) ∘ (sub (liftAList≤ para σ1)))) d
            ≡ fmap G ⟨_⟩ (sub (σ2 +⟨ para ⟩ σ1)) d
    ⟨sub-++-commute⟩′ U t σ1 m1≤m2' m1'≤m2'' para σ2 tt tt = refl
    ⟨sub-++-commute⟩′ I t σ1 m1≤m2' m1'≤m2'' para σ2 d r = r
    ⟨sub-++-commute⟩′ (G₁ ⊕ G₂) t σ1 m1≤m2' m1'≤m2'' para σ2 (inj₁ d) r =
      cong inj₁ (⟨sub-++-commute⟩′ G₁ t σ1 m1≤m2' m1'≤m2'' para σ2 d r)
    ⟨sub-++-commute⟩′ (G₁ ⊕ G₂) t σ1 m1≤m2' m1'≤m2'' para σ2 (inj₂ d) r =
      cong inj₂ (⟨sub-++-commute⟩′ G₂ t σ1 m1≤m2' m1'≤m2'' para σ2 d r)
    ⟨sub-++-commute⟩′ (G₁ ⊗ G₂) t σ1 m1≤m2' m1'≤m2'' para σ2 (d1 , d2) (r1 , r2) =
      cong₂ _,_ (⟨sub-++-commute⟩′ G₁ t σ1 m1≤m2' m1'≤m2'' para σ2 d1 r1)
                (⟨sub-++-commute⟩′ G₂ t σ1 m1≤m2' m1'≤m2'' para σ2 d2 r2)
  
    ⟪sub-++-commute⟫′ : {F : Code} → {m1 m1' m2 m2' m2'' : ℕ} →
            (x : Fin m2'') →
            (σ1 : AList F m1' m1) →
            (m1≤m2' : m1 ≤ m2') →
            (m1'≤m2'' : m1' ≤ m2'') →
            (para : m1'≤m2'' // m1≤m2') →
            (σ2 : AList F m2' m2) →
            cata ⟨_⟩ (sub σ2) (sub (liftAList≤ para σ1) x)
            ≡ sub (σ2 ++ liftAList≤ para σ1) x
    ⟪sub-++-commute⟫′ x σ1 m1≤m2' .m1≤m2' (Refl .m1≤m2') σ2 = refl
    ⟪sub-++-commute⟫′ {F} {m1' = suc l} x σ1 m1≤m2' (s≤s l≤l') (Step para) σ2
      with AListInvAsnoc′ (proj₁ (//→//′ para)) σ1
    ... | (σ , t , y , eq) =
      begin
        cata ⟨_⟩ (sub σ2)
          (((cata ⟨_⟩ (sub (liftAList≤ para σ))) ∘ (lift≤ l≤l' t for inject≤ y (s≤s l≤l'))) x)
      ≡⟨ refl ⟩
        cata ⟨_⟩ (sub σ2)
          ((cata ⟨_⟩ (sub (liftAList≤ para σ))) ((lift≤ l≤l' t for inject≤ y (s≤s l≤l')) x))
      ≡⟨ fuse (sub σ2) (sub (liftAList≤ para σ)) ((lift≤ l≤l' t for inject≤ y (s≤s l≤l')) x) ⟩
        cata ⟨_⟩ (cata ⟨_⟩ (sub σ2) ∘ (sub (liftAList≤ para σ)))
                ((lift≤ l≤l' t for inject≤ y (s≤s l≤l')) x)
      ≡⟨ sub-++-commute′ ((lift≤ l≤l' t for inject≤ y (s≤s l≤l')) x) σ m1≤m2' l≤l' para σ2 ⟩
        (cata ⟨_⟩ (sub (σ2 ++ liftAList≤ para σ))) ((lift≤ l≤l' t for inject≤ y (s≤s l≤l')) x)
      ≡⟨ refl ⟩
        ((cata ⟨_⟩ (sub (σ2 ++ liftAList≤ para σ))) ∘ (lift≤ l≤l' t for inject≤ y (s≤s l≤l'))) x
      ∎
  
    sub-++-commute′ : {F : Code} → {m1 m1' m2 m2' m2'' : ℕ} →
            (t : μ F m2'') →
            (σ1 : AList F m1' m1) →
            (m1≤m2' : m1 ≤ m2') →
            (m1'≤m2'' : m1' ≤ m2'') →
            (para : m1'≤m2'' // m1≤m2') →
            (σ2 : AList F m2' m2) →
            cata ⟨_⟩ (cata ⟨_⟩ (sub σ2) ∘ (sub (liftAList≤ para σ1))) t
            ≡ cata ⟨_⟩ (sub (σ2 +⟨ para ⟩ σ1)) t
    sub-++-commute′ {F} t σ1 m1≤m2' m1'≤m2'' para σ2 =
      ind (λ t → cata ⟨_⟩ (cata ⟨_⟩ (sub σ2) ∘ (sub (liftAList≤ para σ1))) t
                     ≡ cata ⟨_⟩ (sub (σ2 +⟨ para ⟩ σ1)) t)
          (λ d r → cong ⟨_⟩ (⟨sub-++-commute⟩′ F t σ1 m1≤m2' m1'≤m2'' para σ2 d r))
          (λ x → ⟪sub-++-commute⟫′ x σ1 m1≤m2' m1'≤m2'' para σ2)
          t

-- applying empty substitution on t has no effect
applySub≤Anil : {F : Code} → {m : ℕ} → (m≤m : m ≤ m) → (t : μ F m) →
        applySub≤ anil m≤m t ≡ t
applySub≤Anil m≤m t = begin
    applySub≤ anil m≤m t
  ≡⟨ cong (applySub anil) (lift≤-id m≤m t) ⟩
    applySub anil t
  ≡⟨ ⟪-⟫-id t ⟩
    t
  ∎

applyUnifierAppend′ : {F : Code} → {m1 m1' m2 m2' m2'' : ℕ} →
        (f2 : Fin m2' → μ F m2) →
        (σ1 : AList F m1' m1) →
        {m1'≤m2'' : m1' ≤ m2''} →
        {m1≤m2' : m1 ≤ m2'} →
        (para : m1'≤m2'' // m1≤m2') →
        (t : μ F m2'') →
        applyUnifier (f2 +⟨ para ⟩′ sub σ1) t ≡
        applyUnifier f2 (applySub (liftAList≤ para σ1) t)
applyUnifierAppend′ {F} f2 σ1 {m1'≤m2''} {m1≤m2'} para t with //→//′ para
... | (m1≤m1' , m2≤m2' , m1≤m1'//′m2≤m2') = 
  begin
    applyUnifier (applyUnifier f2 ∘ liftFun′ m1≤m1'//′m2≤m2' (sub σ1)) t
  ≡⟨ sym (fuse f2 (liftFun′ m1≤m1'//′m2≤m2' (sub σ1)) t) ⟩
    applyUnifier f2 (applyUnifier (liftFun′ m1≤m1'//′m2≤m2' (sub σ1)) t)
  ≡⟨ cong (applyUnifier f2) (sym (lift-sub-commute′ para m1≤m1'//′m2≤m2'  σ1 t)) ⟩
    applyUnifier f2 (applySub (liftAList≤ para σ1) t)
  ∎

applyUnifierAppend-f′ : {F : Code} → (G : Code) → {m1 m1' m2 m2' m2'' : ℕ} →
        (f2 : Fin m2' → μ F m2) →
        (σ1 : AList F m1' m1) →
        {m1'≤m2'' : m1' ≤ m2''} →
        {m1≤m2' : m1 ≤ m2'} →
        (para : m1'≤m2'' // m1≤m2') →
        (d : ⟦ G ⟧ (μ F m2'')) →
        ⟪fmap⟫ G (f2 +⟨ para ⟩′ sub σ1) d ≡
        ⟪fmap⟫ G f2 (applySub-f G (liftAList≤ para σ1) d)
applyUnifierAppend-f′ {F} G f2 σ1 {m1'≤m2''} {m1≤m2'} para d with //→//′ para
... | (m1≤m1' , m2≤m2' , m1≤m1'//′m2≤m2') =
  begin
    ⟪fmap⟫ G (⟪cata⟫ f2 ∘ liftFun′ m1≤m1'//′m2≤m2' (sub σ1)) d
  ≡⟨ sym (fuse-fmap G f2 (liftFun′ m1≤m1'//′m2≤m2' (sub σ1)) d) ⟩
    ⟪fmap⟫ G f2 (⟪fmap⟫ G (liftFun′ m1≤m1'//′m2≤m2' (sub σ1)) d)
  ≡⟨ cong (⟪fmap⟫ G f2) (sym (lift-sub-commute-f′ G para m1≤m1'//′m2≤m2' σ1 d)) ⟩
    ⟪fmap⟫ G f2 (applySub-f G (liftAList≤ para σ1) d)
  ∎

mutual
  ⟪liftSubCommute⟫ : {F : Code} → (m1 m1' m2' m2'' : ℕ) →
          (σ : AList F m1' m1) →
          (m1≤m2' : m1 ≤ m2') →
          (m1'≤m2'' : m1' ≤ m2'') →
          (para : m1'≤m2'' // m1≤m2') →
          (x : Fin m1') →
          lift≤ m1≤m2' (sub σ x)
          -- cata ⟨_⟩ (λ x → ⟪ inject≤ x m1≤m2' ⟫) (sub σ x)
          ≡ sub (liftAList≤ para σ) (inject≤ x m1'≤m2'')
  ⟪liftSubCommute⟫ {F} m1 .m1 m2' .m2' σ m1≤m2' .m1≤m2' (Refl {m = .m1} {m′ = .m2'} .m1≤m2') x
    with AListInvAnil σ
  ... | σ≡anil rewrite σ≡anil = refl
  ⟪liftSubCommute⟫ {F} m1 (suc l) m2' (suc l') σ m1≤m2' (s≤s l≤l') (Step para) x
    with AListInvAsnoc (s≤s (≤′⇒≤ (//-leq para))) σ
  ... | (.l , refl , σ' , t , y , eq) rewrite eq with thick y x | inspect (thick y) x
  ... | nothing | [ thickyx≡nothing ]
    rewrite thickxy-x≡y y x thickyx≡nothing | thickxx≡nothing (inject≤ x (s≤s l≤l'))
    = liftSubCommute σ' para t
  ... | just x' | [ thickyx≡justx' ] with thickxyjust {x = y} l≤l' thickyx≡justx' 
  ... | eq2 rewrite eq2 = ⟪liftSubCommute⟫ m1 l m2' l' σ' m1≤m2' l≤l' para x'

  liftSubCommute-fmap : {F : Code} → (G : Code) → {m1 m1' m2' m2'' : ℕ} →
          (σ : AList F m1' m1) →
          {m1≤m2' : m1 ≤ m2'} →
          {m1'≤m2'' : m1' ≤ m2''} →
          (para : m1'≤m2'' // m1≤m2') →
          (d : ⟦ G ⟧ (μ F m1')) →
          ⟪fmap⟫ G (⟪inject≤⟫ m1≤m2') (applySub-f G σ d) ≡
          applySub-f≤ G (liftAList≤ para σ) m1'≤m2'' d
  liftSubCommute-fmap U σ para tt = refl
  liftSubCommute-fmap I σ para d = liftSubCommute σ para d
  liftSubCommute-fmap (G₁ ⊕ G₂) σ para (inj₁ d) =
    cong inj₁ (liftSubCommute-fmap G₁ σ para d)
  liftSubCommute-fmap (G₁ ⊕ G₂) σ para (inj₂ d) =
    cong inj₂ (liftSubCommute-fmap G₂ σ para d)
  liftSubCommute-fmap (G₁ ⊗ G₂) σ para (d₁ , d₂) =
    cong₂ _,_ (liftSubCommute-fmap G₁ σ para d₁)
             (liftSubCommute-fmap G₂ σ para d₂)

  liftSubCommute : {F : Code} → {m1 m1' m2' m2'' : ℕ} →
          (σ : AList F m1' m1) →
          {m1≤m2' : m1 ≤ m2'} →
          {m1'≤m2'' : m1' ≤ m2''} →
          (para : m1'≤m2'' // m1≤m2') →
          (t : μ F m1') →
          lift≤ m1≤m2' (applySub σ t) ≡
          applySub≤ (liftAList≤ para σ) m1'≤m2'' t
  liftSubCommute {F} σ para ⟨ d ⟩ =
    cong ⟨_⟩ (liftSubCommute-fmap F σ para d)
  liftSubCommute {F} {m1} {m1'} {m2'} {m2''} σ {m1≤m2'} {m1'≤m2''} para ⟪ x ⟫ = ⟪liftSubCommute⟫ m1 m1' m2' m2'' σ m1≤m2' m1'≤m2'' para x


-- fuse two applySub≤ into one
fuse-applySub≤ : {F : Code} → {m m1 m1' m2 m2' m2'' : ℕ} →
        (t : μ F m) →
        (m≤m1' : m ≤ m1') →
        (σ1 : AList F m1' m1) →
        (m1≤m2' : m1 ≤ m2') →
        (m1'≤m2'' : m1' ≤ m2'') →
        (para : m1'≤m2'' // m1≤m2') →
        (σ2 : AList F m2' m2) →
        (σ : AList F m2''  m2) →
        (m≤m2'' : m ≤ m2'') →
        (σ ≡ σ2 +⟨ para ⟩ σ1) →
        applySub≤ σ m≤m2'' t ≡ applySub≤ σ2 m1≤m2' (applySub≤ σ1 m≤m1' t)
fuse-applySub≤ {F} {m} {m1} {m1'} {m2} {m2'} t m≤m1' σ1 m1≤m2' m1'≤m2'' para σ2 σ m≤m2'' eq
  rewrite eq =
  begin
    applySub≤ (σ2 +⟨ para ⟩ σ1) m≤m2'' t
  ≡⟨ refl ⟩
    applySub (σ2 +⟨ para ⟩ σ1) (lift≤ m≤m2'' t)
  ≡⟨ cong (applySub (σ2 +⟨ para ⟩ σ1)) (sym (fuse-lift m≤m1' m1'≤m2'' m≤m2'' t)) ⟩
    applySub (σ2 +⟨ para ⟩ σ1) (lift≤ m1'≤m2'' (lift≤ m≤m1' t))
  ≡⟨ refl ⟩
    cata ⟨_⟩ (sub (σ2 +⟨ para ⟩ σ1)) (lift≤ m1'≤m2'' (lift≤ m≤m1' t))
  ≡⟨ sym (sub-++-commute′ (lift≤ m1'≤m2'' (lift≤ m≤m1' t)) σ1 m1≤m2' m1'≤m2'' para σ2) ⟩
    cata ⟨_⟩ (cata ⟨_⟩ (sub σ2) ∘ (sub (liftAList≤ para σ1))) (lift≤ m1'≤m2'' (lift≤ m≤m1' t))
  ≡⟨ sym (fuse (sub σ2) (sub (liftAList≤ para σ1)) (lift≤ m1'≤m2'' (lift≤ m≤m1' t))) ⟩
    cata ⟨_⟩ (sub σ2) (cata ⟨_⟩ (sub (liftAList≤ para σ1)) (lift≤ m1'≤m2'' (lift≤ m≤m1' t)))
  ≡⟨ refl ⟩
    applySub σ2 (applySub (liftAList≤ para σ1) (lift≤ m1'≤m2'' (lift≤ m≤m1' t)))
  ≡⟨ cong (applySub σ2) (sym (liftSubCommute σ1 para (lift≤ m≤m1' t))) ⟩
    applySub σ2 (lift≤ m1≤m2' (applySub σ1 (lift≤ m≤m1' t)))
  ≡⟨ refl ⟩
    applySub≤ σ2 m1≤m2' (applySub≤ σ1 m≤m1' t)
  ∎

fuse-applyUnifier′≤ : {F : Code} → {m m′ m′′ k′′ k′′′ l′′′ l′ : ℕ} →
        (f : Fin l′′′ → μ F l′) →
        (f′ : Fin k′′ → μ F l′) →
        (σ : AList F m′′ m′) →
        (m′′≤k′′′ : m′′ ≤ k′′′) →
        (m′≤k′′ : m′ ≤ k′′) →
        (para : m′′≤k′′′ // m′≤k′′) →
        (m≤l′′′ : m ≤ l′′′) →
        (m≤m′′ : m ≤ m′′) →
        (m≤k′′′ : m ≤ k′′′) →
        ((x : Fin m) →
         f (inject≤ x m≤l′′′) ≡
         (f′ +⟨ para ⟩′ sub σ) (inject≤ x m≤k′′′)) →
        (t : μ F m) →
        applyUnifier≤ f m≤l′′′ t ≡
        applyUnifier≤ f′ m′≤k′′ (applySub≤ σ m≤m′′ t)
fuse-applyUnifier′≤ {F} f f′ σ m′′≤k′′′ m′≤k′′ para m≤l′′′ m≤m′′ m≤k′′′ eq t =
  begin
    applyUnifier≤ f m≤l′′′ t
  ≡⟨ cata-ext′′ f (f′ +⟨ para ⟩′ sub σ) m≤l′′′ m≤k′′′ eq t ⟩
    applyUnifier≤ (f′ +⟨ para ⟩′ sub σ) m≤k′′′ t
  ≡⟨ refl ⟩
    applyUnifier (f′ +⟨ para ⟩′ sub σ) (lift≤ m≤k′′′ t)
  ≡⟨ cong (applyUnifier (f′ +⟨ para ⟩′ sub σ))
          (sym (fuse-lift m≤m′′ m′′≤k′′′ m≤k′′′ t)) ⟩
    applyUnifier (f′ +⟨ para ⟩′ sub σ) (lift≤ m′′≤k′′′ (lift≤ m≤m′′ t))
  ≡⟨ applyUnifierAppend′ f′ σ para (lift≤ m′′≤k′′′ (lift≤ m≤m′′ t)) ⟩
    applyUnifier f′ (applySub (liftAList≤ para σ)
                      (lift≤ m′′≤k′′′ (lift≤ m≤m′′ t)))
  ≡⟨ cong (applyUnifier f′)
          (sym (liftSubCommute σ para (lift≤ m≤m′′ t))) ⟩
    applyUnifier f′ (lift≤ m′≤k′′ (applySub σ (lift≤ m≤m′′ t)))
  ≡⟨ refl ⟩
    applyUnifier≤ f′ m′≤k′′ (applySub≤ σ m≤m′′ t)
  ∎

para-sub-append : {F : Code} → {m′ m′′ m′′′ m₁ m₂ k₁′′ k₂′′ : ℕ} →
        (f₂′ : Fin k₂′′ → μ F m′′) →
        {m₁≤k₁′′ : m₁ ≤ k₁′′} → {m₂≤k₂′′ : m₂ ≤ k₂′′} →
        (para₂ :  m₁≤k₁′′ // m₂≤k₂′′) →
        (σ₂ : AList F m₁ m₂) →
        {m′≤m′′′ : m′ ≤ m′′′} →
        (para₁ :  m′≤m′′′ // m₁≤k₁′′) →
        (σ₁ : AList F m′ m₁) →
        (x : Fin m′) →
        ((f₂′ +⟨ para₂ ⟩′ (sub σ₂)) +⟨ para₁ ⟩′ (sub σ₁)) (inject≤ x m′≤m′′′) ≡
        (f₂′ +⟨ para-append para₁ para₂ ⟩′ (sub (σ₂ ++ σ₁))) (inject≤ x m′≤m′′′)
para-sub-append {F} f₂′ {m₁≤k₁′′} {m₂≤k₂′′} para₂ σ₂ {m′≤m′′′} para₁ σ₁ x =
  begin
    ((f₂′ +⟨ para₂ ⟩′ (sub σ₂)) +⟨ para₁ ⟩′ (sub σ₁)) (inject≤ x m′≤m′′′)
  ≡⟨ refl ⟩
    applyUnifier ((f₂′ +⟨ para₂ ⟩′ (sub σ₂)) +⟨ para₁ ⟩′ (sub σ₁))
         ⟪ inject≤ x m′≤m′′′ ⟫
  ≡⟨ applyUnifierAppend′ (f₂′ +⟨ para₂ ⟩′ (sub σ₂)) σ₁ para₁ ⟪ inject≤ x m′≤m′′′ ⟫ ⟩
    applyUnifier (applyUnifier (f₂′ +⟨ para₂ ⟩′ (sub σ₂)) ∘
                   (sub (liftAList≤ para₁ σ₁)))
         ⟪ inject≤ x m′≤m′′′ ⟫
  ≡⟨ refl ⟩
    applyUnifier (f₂′ +⟨ para₂ ⟩′ (sub σ₂))
      (sub (liftAList≤ para₁ σ₁)
         (inject≤ x m′≤m′′′))
  ≡⟨ applyUnifierAppend′ f₂′ σ₂ para₂
      (sub (liftAList≤ para₁ σ₁) (inject≤ x m′≤m′′′)) ⟩
    applyUnifier f₂′ (applySub (liftAList≤ para₂ σ₂)
      (sub (liftAList≤ para₁ σ₁)
         (inject≤ x m′≤m′′′)))
  ≡⟨ refl ⟩
    applyUnifier f₂′ (applySub (liftAList≤ para₂ σ₂)
      (applySub (liftAList≤ para₁ σ₁)
         (lift≤ m′≤m′′′ ⟪ x ⟫)))
  ≡⟨ cong (λ t →
           applyUnifier f₂′
           (applySub (liftAList≤ para₂ σ₂) t))
        (sym (liftSubCommute σ₁ para₁ ⟪ x ⟫)) ⟩
    applyUnifier f₂′ (applySub (liftAList≤ para₂ σ₂)
      (lift≤ m₁≤k₁′′ (applySub σ₁ ⟪ x ⟫)))
  ≡⟨ cong (applyUnifier f₂′) (sym (liftSubCommute σ₂ para₂ (applySub σ₁ ⟪ x ⟫))) ⟩
    applyUnifier f₂′ (lift≤ m₂≤k₂′′ (applySub σ₂ (applySub σ₁ ⟪ x ⟫)))
  ≡⟨ refl ⟩
    applyUnifier f₂′ (lift≤ m₂≤k₂′′
      (applyUnifier (applySub σ₂ ∘ (sub σ₁)) ⟪ x ⟫))
  ≡⟨ cong (λ t → applyUnifier f₂′ (lift≤ m₂≤k₂′′ t))
       (sub-++-commute σ₂ σ₁ ⟪ x ⟫) ⟩
    applyUnifier f₂′ (lift≤ m₂≤k₂′′ (applySub (σ₂ ++ σ₁) ⟪ x ⟫))
  ≡⟨ cong (applyUnifier f₂′)
       (liftSubCommute (σ₂ ++ σ₁) (para-append para₁ para₂) ⟪ x ⟫) ⟩
    applyUnifier f₂′ (applySub (liftAList≤ (para-append para₁ para₂) (σ₂ ++ σ₁))
       (lift≤ m′≤m′′′ ⟪ x ⟫))
  ≡⟨ refl ⟩
    applyUnifier f₂′ (applySub (liftAList≤ (para-append para₁ para₂) (σ₂ ++ σ₁))
         ⟪ inject≤ x m′≤m′′′ ⟫)
  ≡⟨ sym (applyUnifierAppend′ f₂′ (σ₂ ++ σ₁) (para-append para₁ para₂) ⟪ inject≤ x m′≤m′′′ ⟫) ⟩
    applyUnifier (f₂′ +⟨ para-append para₁ para₂ ⟩′ (sub (σ₂ ++ σ₁)))
         ⟪ inject≤ x m′≤m′′′ ⟫
  ≡⟨ refl ⟩
    (f₂′ +⟨ para-append para₁ para₂ ⟩′ (sub (σ₂ ++ σ₁))) (inject≤ x m′≤m′′′)
  ∎

private
  applySubC : {F : Code} → (G : Code) → {m m′ : ℕ} →
           (f′ : Fin m → μ F m′) →
           (f : Ctx⟦ G ⟧ (μ F m)) → Ctx⟦ G ⟧ (μ F m′)
  applySubC U f′ ()
  applySubC I f′ f = f
  applySubC (G₁ ⊕ G₂) f′ (inj₁ f) = inj₁ (applySubC G₁ f′ f)
  applySubC (G₁ ⊕ G₂) f′ (inj₂ f) = inj₂ (applySubC G₂ f′ f)
  applySubC {F} (G₁ ⊗ G₂) f′ (inj₁ (f₁ , d₂)) =
    inj₁ (applySubC G₁ f′ f₁ , fmap G₂ ⟨_⟩ f′ d₂)
  applySubC {F} (G₁ ⊗ G₂) f′ (inj₂ (d₁ , f₂)) =
    inj₂ (fmap G₁ ⟨_⟩ f′ d₁ , applySubC G₂ f′ f₂)

  applySubC≤ : {F : Code} → (G : Code) → {m m′ m′′′ : ℕ} →
           (f′ : Fin m′′′ → μ F m′) →
           (m≤m′′′ : m ≤ m′′′) →
           (f : Ctx⟦ G ⟧ (μ F m)) → Ctx⟦ G ⟧ (μ F m′)
  applySubC≤ U f′ m≤m′′′ ()
  applySubC≤ I f′ m≤m′′′ f = f
  applySubC≤ (G₁ ⊕ G₂) f′ m≤m′′′ (inj₁ f) = inj₁ (applySubC≤ G₁ f′ m≤m′′′ f)
  applySubC≤ (G₁ ⊕ G₂) f′ m≤m′′′ (inj₂ f) = inj₂ (applySubC≤ G₂ f′ m≤m′′′ f)
  applySubC≤ {F} (G₁ ⊗ G₂) f′ m≤m′′′ (inj₁ (f₁ , d₂)) =
    inj₁ (applySubC≤ G₁ f′ m≤m′′′ f₁ ,
         fmap G₂ ⟨_⟩ (λ x → f′ (inject≤ x m≤m′′′)) d₂)
  applySubC≤ {F} (G₁ ⊗ G₂) f′ m≤m′′′ (inj₂ (d₁ , f₂)) =
    inj₂ (fmap G₁ ⟨_⟩ (λ x → f′ (inject≤ x m≤m′′′)) d₁ ,
         applySubC≤ G₂ f′ m≤m′′′ f₂)

applySubCs : {F : Code} → {m m′ : ℕ} →
           (f′ : Fin m → μ F m′) →
           (fs : Context F (μ F m)) → Context F (μ F m′)
applySubCs f′ [] = []
applySubCs {F} f′ (f ∷ fs) = applySubC F f′ f ∷ applySubCs f′ fs

applySubCs≤ : {F : Code} → {m m′ m′′′ : ℕ} →
           (f′ : Fin m′′′ → μ F m′) →
           (m≤m′′′ : m ≤ m′′′) →
           (fs : Context F (μ F m)) → Context F (μ F m′)
applySubCs≤ f′ m≤m′′′ [] = []
applySubCs≤ {F} f′ m≤m′′′ (f ∷ fs) =
  applySubC≤ F f′ m≤m′′′ f ∷ applySubCs≤ f′ m≤m′′′ fs

private
  fmapSubPlug : {F : Code} → (G : Code) → {m m′ : ℕ} →
        (f′ : Fin m → μ F m′) →
        (f : Ctx⟦ G ⟧ (μ F m)) →
        (t : μ F m) → 
        fmap G ⟨_⟩ f′ (plug′ G f t) ≡
        plug′ G (applySubC G f′ f) (applyUnifier f′ t)
  fmapSubPlug U f′ () t
  fmapSubPlug I f′ tt t = refl
  fmapSubPlug (G₁ ⊕ G₂) f′ (inj₁ f) t = cong inj₁ (fmapSubPlug G₁ f′ f t)
  fmapSubPlug (G₁ ⊕ G₂) f′ (inj₂ f) t = cong inj₂ (fmapSubPlug G₂ f′ f t)
  fmapSubPlug (G₁ ⊗ G₂) f′ (inj₁ (f₁ , d₂)) t =
    cong₂ _,_ (fmapSubPlug G₁ f′ f₁ t) refl
  fmapSubPlug (G₁ ⊗ G₂) f′ (inj₂ (d₁ , f₂)) t =
    cong₂ _,_ refl (fmapSubPlug G₂ f′ f₂ t)

  fmapSubPlug≤ : {F : Code} → (G : Code) → {m m′ m′′′ : ℕ} →
        (f′ : Fin m′′′ → μ F m′) →
        (m≤m′′′ : m ≤ m′′′) →
        (f : Ctx⟦ G ⟧ (μ F m)) →
        (t : μ F m) → 
        fmap G ⟨_⟩ (λ x → f′ (inject≤ x m≤m′′′)) (plug′ G f t) ≡
        plug′ G (applySubC≤ G f′ m≤m′′′ f) (applyUnifier≤ f′ m≤m′′′ t)
  fmapSubPlug≤ U f′ m≤m′′′ () t
  fmapSubPlug≤ I f′ m≤m′′′ tt t = sym (fuse f′ (λ x → ⟪ inject≤ x m≤m′′′ ⟫) t)
  fmapSubPlug≤ (G₁ ⊕ G₂) f′ m≤m′′′ (inj₁ f) t = cong inj₁ (fmapSubPlug≤ G₁ f′ m≤m′′′ f t)
  fmapSubPlug≤ (G₁ ⊕ G₂) f′ m≤m′′′ (inj₂ f) t = cong inj₂ (fmapSubPlug≤ G₂ f′ m≤m′′′ f t)
  fmapSubPlug≤ (G₁ ⊗ G₂) f′ m≤m′′′ (inj₁ (f₁ , d₂)) t =
    cong₂ _,_ (fmapSubPlug≤ G₁ f′ m≤m′′′ f₁ t) refl
  fmapSubPlug≤ (G₁ ⊗ G₂) f′ m≤m′′′ (inj₂ (d₁ , f₂)) t =
    cong₂ _,_ refl (fmapSubPlug≤ G₂ f′ m≤m′′′ f₂ t)

applySubPlug : {F : Code} → {m m′ : ℕ} →
        (f′ : Fin m → μ F m′) →
        (fs : Context F (μ F m)) →
        (t : μ F m) → 
        applyUnifier f′ (plug fs t) ≡
        plug (applySubCs f′ fs) (applyUnifier f′ t)
applySubPlug f′ [] t = refl
applySubPlug {F} f′ (f ∷ fs) t =
  begin
    ⟨ fmap F ⟨_⟩ f′ (plug′ F f (plug fs t)) ⟩
  ≡⟨ cong ⟨_⟩ (fmapSubPlug F f′ f (plug fs t)) ⟩
    ⟨ plug′ F (applySubC F f′ f) (applyUnifier f′ (plug fs t)) ⟩
  ≡⟨ cong (λ t → ⟨ plug′ F (applySubC F f′ f) t ⟩) (applySubPlug f′ fs t) ⟩
    ⟨ plug′ F (applySubC F f′ f) (plug (applySubCs f′ fs) (applyUnifier f′ t)) ⟩
  ∎

applySubPlug≤ : {F : Code} → {m m′ m′′′ : ℕ} →
        (f′ : Fin m′′′ → μ F m′) →
        (m≤m′′′ : m ≤ m′′′) →
        (fs : Context F (μ F m)) →
        (t : μ F m) → 
        applyUnifier≤ f′ m≤m′′′ (plug fs t) ≡
        plug (applySubCs≤ f′ m≤m′′′ fs) (applyUnifier≤ f′ m≤m′′′ t)
applySubPlug≤ f′ m≤m′′′ [] t = refl
applySubPlug≤ {F} f′ m≤m′′′ (f ∷ fs) t =
  begin
    applyUnifier≤ f′ m≤m′′′ ⟨ plug′ F f (plug fs t) ⟩
  ≡⟨ fuse f′ (λ x → ⟪ inject≤ x m≤m′′′ ⟫) ⟨ plug′ F f (plug fs t) ⟩ ⟩
    ⟨ fmap F ⟨_⟩ (λ x → f′ (inject≤ x m≤m′′′)) (plug′ F f (plug fs t)) ⟩
  ≡⟨ cong ⟨_⟩ (fmapSubPlug≤ F f′ m≤m′′′ f (plug fs t)) ⟩
    ⟨ plug′ F (applySubC≤ F f′ m≤m′′′ f) (applyUnifier≤ f′ m≤m′′′ (plug fs t)) ⟩
  ≡⟨ cong (λ t → ⟨ plug′ F (applySubC≤ F f′ m≤m′′′ f) t ⟩)
          (applySubPlug≤ f′ m≤m′′′ fs t) ⟩
    ⟨ plug′ F (applySubC≤ F f′ m≤m′′′ f) (plug (applySubCs≤ f′ m≤m′′′ fs) (applyUnifier≤ f′ m≤m′′′ t)) ⟩
  ∎

private
  ⟪for-thin-id⟫′ : (F : Code) → {m : ℕ} → (x : Fin (suc m)) → (y : Fin m) → (t′ : μ F m) →
          (applyUnifier (t′ for x) ∘ (⟪_⟫ ∘ thin x)) y ≡ ⟪ y ⟫
  ⟪for-thin-id⟫′ F zero y t′ = refl
  ⟪for-thin-id⟫′ F {zero} (suc ()) y t′
  ⟪for-thin-id⟫′ F {suc m} (suc x) zero t′ = refl
  ⟪for-thin-id⟫′ F {suc m} (suc x) (suc y) t′
    with thick x (suc y) | inspect (thick x) (suc y)
  ... | nothing | [ eq ]
    rewrite thickxy-x≡y x (suc y) eq | thick-thin-eq (suc y) y = refl
  ... | just y' | [ eq ]
    rewrite thick-thin-eq x y = refl

  for-thin-id′ : (F : Code) → {m : ℕ} → (x : Fin (suc m)) →
          (t : μ F m) → (t′ : μ F m) →
          applyUnifier (applyUnifier (t′ for x) ∘ (⟪_⟫ ∘ thin x)) t ≡ t
  for-thin-id′ F x t t′ =
    cata-id (applyUnifier (t′ for x) ∘ (⟪_⟫ ∘ thin x))
            (λ y → ⟪for-thin-id⟫′ F x y t′)
            t

for-thin′ : (F : Code) → {m : ℕ} → (x : Fin (suc m)) → (t : μ F m) → (t′ : μ F m) →
        applyUnifier (t′ for x) (applyUnifier (⟪_⟫ ∘ thin x) t) ≡ t
for-thin′ F x t t′ =
  begin
    applyUnifier (t′ for x) (applyUnifier (⟪_⟫ ∘ thin x) t)
  ≡⟨ fuse (t′ for x) (⟪_⟫ ∘ thin x) t ⟩
    applyUnifier (applyUnifier (t′ for x) ∘ (⟪_⟫ ∘ thin x)) t
  ≡⟨ for-thin-id′ F x t t′ ⟩
    t
  ∎

-- most general unifier

mg : {F : Code} → {m m′ : ℕ} → (t₁ t₂ : μ F m) →
        (g : Fin m → μ F m′) → Set
mg {F} {m} {m′} t₁ t₂ g =
  (⟪cata⟫ g t₁ ≡ ⟪cata⟫ g t₂) ×
  ((l′ l′′ : ℕ) → (f : Fin l′′ → μ F l′) → (m≤l′′ : m ≤ l′′) →
   (⟪cata⟫≤ f m≤l′′ t₁ ≡ ⟪cata⟫≤ f m≤l′′ t₂) → 
   (Σ[ k′′ ∈ ℕ ] Σ[ m′≤k′′ ∈ m′ ≤ k′′ ] Σ[ para ∈ m≤l′′ // m′≤k′′ ]
    Σ[ f′ ∈ (Fin k′′ → μ F l′) ]
    ((x : Fin m) →
     f (inject≤ x m≤l′′) ≡ (f′ +⟨ para ⟩′ g) (inject≤ x m≤l′′))))

mg-f : {F : Code} → (G : Code) → {m m′ : ℕ} →
        (d₁ d₂ : ⟦ G ⟧ (μ F m)) → (g : Fin m → μ F m′) → Set
mg-f {F} G {m} {m′} d₁ d₂ g =
  (⟪fmap⟫ G g d₁ ≡ ⟪fmap⟫ G g d₂) ×
  ((l′ l′′ : ℕ) → (f : Fin l′′ → μ F l′) → (m≤l′′ : m ≤ l′′) →
   (⟪fmap⟫≤ G f m≤l′′ d₁ ≡ ⟪fmap⟫≤ G f m≤l′′ d₂) → 
   (Σ[ k′′ ∈ ℕ ] Σ[ m′≤k′′ ∈ m′ ≤ k′′ ] Σ[ para ∈ m≤l′′ // m′≤k′′ ]
    Σ[ f′ ∈ (Fin k′′ → μ F l′) ]
    ((x : Fin m) →
     f (inject≤ x m≤l′′) ≡ (f′ +⟨ para ⟩′ g) (inject≤ x m≤l′′))))

mgAnil : {F : Code} → {m : ℕ} → (t : μ F m) →
        mg t t ⟪_⟫
mgAnil _ =
  (refl , λ l′ l′′ f′ m≤l′′ _ →
            (l′′ , m≤l′′ , Refl m≤l′′ ,
             f′ , (λ x → sym (append-nil′ m≤l′′ f′ (inject≤ x m≤l′′)))))

mgAnil-f : {F : Code} → (G : Code) → {m : ℕ} → (d : ⟦ G ⟧ (μ F m)) →
        mg-f G d d ⟪_⟫
mgAnil-f G _ =
  (refl , λ l′ l′′ f m≤l′′ _ →
            (l′′ , m≤l′′ , Refl m≤l′′ , f ,
            λ x → sym (append-nil′ m≤l′′ f (inject≤ x m≤l′′))))

private
  thin-for≤′ : {F : Code} → {m l′ l′′ : ℕ} →
          (x y : Fin (suc m)) →
          (f : Fin (suc l′′) → μ F l′) →
          (t′ : μ F m) →
          (m≤l′′ : m ≤ l′′) →
          (eq : f (inject≤ x (s≤s m≤l′′)) ≡
                applyUnifier≤ f (s≤s m≤l′′)
                 (applyUnifier (⟪_⟫ ∘ thin x) t′)) →
          f (inject≤ y (s≤s m≤l′′)) ≡
          applyUnifier (f ∘ thin (inject≤ x (s≤s m≤l′′)))
            (lift≤ m≤l′′ ((t′ for x) y))
  thin-for≤′ {F} x y f t′ m≤l′′ eq with thick x y | inspect (thick x) y
  ... | nothing | [ eq′ ]
    rewrite sym (thickxy-x≡y x y eq′) =
    begin
      f (inject≤ x (s≤s m≤l′′))
    ≡⟨ eq ⟩
      applyUnifier f
        (lift≤ (s≤s m≤l′′) (applyUnifier (⟪_⟫ ∘ thin x) t′))
    ≡⟨ cong (applyUnifier f)
        (fuse (λ x → ⟪ inject≤ x (s≤s m≤l′′) ⟫)
              (⟪_⟫ ∘ thin x) t′) ⟩
      applyUnifier f
        (applyUnifier
          (applyUnifier (λ x → ⟪ inject≤ x (s≤s m≤l′′) ⟫) ∘
           (⟪_⟫ ∘ thin x))
          t′)
    ≡⟨ cong (applyUnifier f)
         (cata-ext
           (applyUnifier (λ x → ⟪ inject≤ x (s≤s m≤l′′) ⟫) ∘
            (⟪_⟫ ∘ thin x))
           (applyUnifier (⟪_⟫ ∘ thin (inject≤ x (s≤s m≤l′′))) ∘
            (λ x → ⟪ inject≤ x m≤l′′ ⟫))
           (λ y → cong ⟪_⟫ (inject-thin x y m≤l′′))
           t′) ⟩
      applyUnifier f
        (applyUnifier 
          (applyUnifier (⟪_⟫ ∘ thin (inject≤ x (s≤s m≤l′′))) ∘
           (λ x → ⟪ inject≤ x m≤l′′ ⟫))
          t′)
    ≡⟨ cong (applyUnifier f)
        (sym (fuse (⟪_⟫ ∘ thin (inject≤ x (s≤s m≤l′′)))
                   (λ x → ⟪ inject≤ x m≤l′′ ⟫) t′)) ⟩
      applyUnifier f
        (applyUnifier (⟪_⟫ ∘ thin (inject≤ x (s≤s m≤l′′)))
          (applyUnifier (λ x → ⟪ inject≤ x m≤l′′ ⟫) t′))
    ≡⟨ fuse f (⟪_⟫ ∘ thin (inject≤ x (s≤s m≤l′′)))
            (applyUnifier (λ x → ⟪ inject≤ x m≤l′′ ⟫) t′) ⟩
      applyUnifier (f ∘ thin (inject≤ x (s≤s m≤l′′)))
        (applyUnifier (λ x → ⟪ inject≤ x m≤l′′ ⟫) t′)
    ≡⟨ refl ⟩
      applyUnifier (f ∘ thin (inject≤ x (s≤s m≤l′′))) (lift≤ m≤l′′ t′)
    ∎
  ... | just y′ | [ eq′ ]
    rewrite thick-thin (inject≤ x (s≤s m≤l′′)) (inject≤ y (s≤s m≤l′′)) (inject≤ y′ m≤l′′) (thickxyjust {x = x} {y = y} m≤l′′ eq′) = refl

private
  h : {F : Code} → {m : ℕ} →
   (t′ : μ F m) → (x : Fin (suc m)) →
   (l′ l′′ : ℕ) → (f′ : Fin l′′ → μ F l′) → (1+m≤l′′ : suc m ≤ l′′) →
   (f′ (inject≤ x 1+m≤l′′) ≡
    applyUnifier≤ f′ 1+m≤l′′
      (applyUnifier (λ y → ⟪ thin x y ⟫) t′)) → 
   (Σ[ k′′ ∈ ℕ ]
    Σ[ m≤k′′ ∈ m ≤ k′′ ]
    Σ[ para ∈ 1+m≤l′′ // m≤k′′ ]
    Σ[ f′′ ∈ (Fin k′′ → μ F l′) ]
    ((y : Fin (suc m)) →
     f′ (inject≤ y 1+m≤l′′) ≡
     (f′′ +⟨ para ⟩′ (sub (anil asnoc t′ / x))) (inject≤ y 1+m≤l′′)))
  h {F} {m} t′ x l′ .(suc l′′) f′ (s≤s {m = .m} {n = l′′} m≤l′′) eq
    with //→//′ (Step (Refl m≤l′′))
  ... | (m≤1+m , l′′≤1+l′′ , m≤1+m//′l′′≤1+l′′) =
    (l′′ , m≤l′′ , Step (Refl m≤l′′) ,
     f′ ∘ thin (inject≤ x (s≤s m≤l′′)) ,
     λ y →
        begin
          f′ (inject≤ y (s≤s m≤l′′))
        ≡⟨ thin-for≤′ x y f′ t′ m≤l′′ eq ⟩
          applyUnifier (f′ ∘ thin (inject≤ x (s≤s m≤l′′)))
           (lift≤ m≤l′′ ((t′ for x) y))
        ≡⟨ cong (λ t → 
              applyUnifier (f′ ∘ thin (inject≤ x (s≤s m≤l′′)))
               (lift≤ m≤l′′ t))
              (sym (⟪-⟫-id ((t′ for x) y))) ⟩
          applyUnifier (f′ ∘ thin (inject≤ x (s≤s m≤l′′)))
           (lift≤ m≤l′′ (applyUnifier ⟪_⟫ ((t′ for x) y)))
        ≡⟨ refl ⟩
          applyUnifier (f′ ∘ thin (inject≤ x (s≤s m≤l′′)))
           (lift≤ m≤l′′ (applyUnifier (cata ⟨_⟩ ⟪_⟫ ∘ (t′ for x)) ⟪ y ⟫))
        ≡⟨ refl ⟩
          applyUnifier (f′ ∘ thin (inject≤ x (s≤s m≤l′′)))
           (lift≤ m≤l′′ (applySub (anil asnoc t′ / x) ⟪ y ⟫))
        ≡⟨ cong (applyUnifier (f′ ∘ thin (inject≤ x (s≤s m≤l′′))))
                (liftSubCommute (anil asnoc t′ / x) (Step (Refl m≤l′′)) ⟪ y ⟫) ⟩
          applyUnifier (f′ ∘ thin (inject≤ x (s≤s m≤l′′)))
           (applySub (liftAList≤ (Step (Refl m≤l′′))
                                 (anil asnoc t′ / x))
                     (lift≤ (s≤s m≤l′′) ⟪ y ⟫))
        ≡⟨ sym (applyUnifierAppend′ (f′ ∘ thin (inject≤ x (s≤s m≤l′′))) (anil asnoc t′ / x) (Step (Refl m≤l′′)) ⟪ inject≤ y (s≤s m≤l′′) ⟫) ⟩
          applyUnifier ((f′ ∘ thin (inject≤ x (s≤s m≤l′′)))
           +⟨ Step (Refl m≤l′′) ⟩′ (sub (anil asnoc t′ / x)))
          ⟪ inject≤ y (s≤s m≤l′′) ⟫
        ≡⟨ refl ⟩
          ((f′ ∘ thin (inject≤ x (s≤s m≤l′′)))
           +⟨ Step (Refl m≤l′′) ⟩′ (sub (anil asnoc t′ / x)))
          (inject≤ y (s≤s m≤l′′))
        ∎)

mgFor : {F : Code} → {m : ℕ} →
        (x : Fin (suc m)) → (t′ : μ F m) →
        mg ⟪ x ⟫ (applyUnifier (⟪_⟫ ∘ thin x) t′)
                     (sub (anil asnoc t′ / x))
mgFor {F} {m} x t′
  rewrite thickxx≡nothing x =
  ((begin applyUnifier ⟪_⟫ t′
        ≡⟨ ⟪-⟫-id t′ ⟩
          t′
        ≡⟨ sym (for-thin′ F x t′ t′) ⟩
          applyUnifier (t′ for x) (applyUnifier (⟪_⟫ ∘ thin x) t′)
        ≡⟨ sym (⟪-⟫-id (applyUnifier (t′ for x) (applyUnifier (⟪_⟫ ∘ thin x) t′))) ⟩
          applyUnifier ⟪_⟫ (applyUnifier (t′ for x) (applyUnifier (⟪_⟫ ∘ thin x) t′))
        ≡⟨ fuse ⟪_⟫ (t′ for x) (applyUnifier (⟪_⟫ ∘ thin x) t′) ⟩
          applyUnifier (applyUnifier ⟪_⟫ ∘ t′ for x) (applyUnifier (⟪_⟫ ∘ thin x) t′)
        ∎) ,
   (λ l′ l′′ f′ 1+m≤l′′ eq → h t′ x l′ l′′ f′ 1+m≤l′′ eq ))
